#ifndef _DAQDATAFORMATEXCEPTION_H
#define _DAQDATAFORMATEXCEPTION_H
#include "DAQDataFormats/RawDAQData.h" //< using DEBUGGING variable from there. Otherwise, we don't need it

#include <stdexcept>
#include <execinfo.h>
#include <sstream>
#include <cstdlib>

#ifdef DEBUGGING
#define GENERATE_DATAFORMAT_EXCEPTION(msg) \
  throw daqdataformats::DataFormatException(__PRETTY_FUNCTION__, __FILE__, __LINE__, msg, _dataFormatClassName, _dataFormatConstructor )
#else
#define GENERATE_DATAFORMAT_EXCEPTION(msg) \
  throw daqdataformats::DataFormatException(__PRETTY_FUNCTION__, __FILE__, __LINE__, msg )
#endif

#define GENERATE_DATAFORMAT_WRONG_VERSION_EXCEPTION GENERATE_DATAFORMAT_EXCEPTION(generateExceptionForWrongVersion(MAXIMUM_VERSION))

namespace daqdataformats {

/**
 * Class to report exceptions in the DAQDataFormats package.
 *
 * @author Kurt Biery
 * @version $Revision: 1.7 $ $Date: 2013/10/23 15:24:39 $
 */

class DataFormatException : public std::exception {

  public:
    
    /**
     * Constructs a data format exception instance from the specified
     * file, line number, and message.
     */
    DataFormatException( const std::string& pretty_function
                        ,const std::string& file_name
                        ,int                line
                        ,const std::string& message
                        ,const std::string  dataformat_class_name  = ""
                        ,const std::string  dataformat_constructor = ""
                       ):
          _pretty_function       (&pretty_function)
         ,_file_name             (&file_name)
         ,_line                  (line)
         ,_message               (&message)
         ,_dataformat_class_name (&dataformat_class_name)
         ,_dataformat_constructor(&dataformat_constructor)
    {
             constructMessage();
    }// end of constructor


    DataFormatException(const std::string& message):
          _pretty_function       (NULL)
         ,_file_name             (NULL)
         ,_line                  (0)
         ,_message               (&message)
         ,_dataformat_class_name (NULL)
         ,_dataformat_constructor(NULL)
    {
             constructMessage();
    }// end of constructor

    virtual ~DataFormatException() throw() { }
    /**
     * Returns a description of the exception.
     */
    const char* what() const throw()
    {
        return _fullText.c_str();
    }

 private:
    //--------------------------------------------------------------------------
    /// Construct exception message
    bool constructMessage()
    {
            /// build up the exception text
            _fullText  = *_message + "\n" + " DataFormatException generated";

            /// Add function name
            if(_pretty_function && _pretty_function->size() > 0){
                   _fullText.append(" in function" + *_pretty_function);
            }

            /// Add File name
            if(_file_name && _file_name->size() > 0){
                _fullText += " in File " + *_file_name;
            }

            /// Add Line number
            if(_line > 0){
                std::stringstream line_stream;
                line_stream<<_line;
                _fullText += " Line:" + line_stream.str() + ".";
            }

            /// if dataformat class name is given, print it
            if(_dataformat_class_name && _dataformat_class_name->size() > 0){
                _fullText += "\nDataFormat Class: "       + *_dataformat_class_name + ".";
            }

            /// if dataformat constructor name is given, print it
            if(_dataformat_constructor && _dataformat_constructor->size() > 0){
                _fullText += " DataFormat Constructor: "  + *_dataformat_constructor + ".";
            }// end if dataformat class name is given

            _fullText.append(".");

            /// add some limited stack trace information
            void *array[10];
            const size_t size = backtrace (array, 10);

            char** strings = backtrace_symbols (array, size);

            std::stringstream tmpStream;
            for (size_t idx = 0; idx < size; idx++) {
                tmpStream << "      " << strings[idx] << std::endl;
            }
            free (strings);

            _fullText += "\n    Stack trace:\n" + tmpStream.str();

            //std::cout<<_fullText<<"\n";

            return true;
    }

 private:
         const std::string*  _pretty_function;
         const std::string*  _file_name;
         int                 _line;
         const std::string*  _message;
         const std::string*  _dataformat_class_name;
         const std::string*  _dataformat_constructor;

    /**
     * The full description of the exception.
     */
    std::string _fullText;

};

};  // end of namespace

#endif
