#ifndef _DBUTILS_DAQCONFIG_CONFIGDATATREE_H
#define _DBUTILS_DAQCONFIG_CONFIGDATATREE_H

#include <DatabaseUtils/DAQConfig/ConfigDataSelector.h>
#include <DatabaseUtils/SimpleParameterSet.h>
#include <NovaDatabase/Table.h>
#include <boost/algorithm/string/replace.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/tuple/tuple.hpp>
#include <string>
#include <map>
#include <vector>
#include <iostream>
#include <stdint.h>

namespace dbutils {
namespace daqconfig {

/**
 * Class to handle fetching of DAQ configuration data from the database.
 *
 * @author Kurt Biery
 * @version $Revision: 1.12 $ $Date: 2011/08/22 14:10:22 $
 */

class ConfigDataTree {

 public:

    static const std::string KEY_FOR_SYSTEM_WIDE_PARAM;
    static const std::string KEY_SEPARATOR_STRING;
    static const std::string NO_DEFAULT_KEY_STRING;

    typedef std::map<std::string, std::string> ConfigValueMap;
    typedef boost::shared_ptr<ConfigValueMap> ConfigValueMapPtr;
    typedef std::map<std::string, ConfigValueMapPtr> ConfigDataContainer;
    typedef std::vector<std::string> ConfigValueList;
    typedef boost::shared_ptr<ConfigValueList> ConfigValueListPtr;
    typedef std::map<std::string, ConfigValueListPtr> ConfigValueListMap;

    // **********************
    // *** Constructor(s) ***
    // **********************

    ConfigDataTree(const std::string& tableDefinitionFilename,
                   const std::vector<int64_t>& subsystemConfigIdList,
                   const ConfigDataSelector& selector,
                   std::vector<std::string> dataColumnNameList);

    ConfigDataTree(const std::string& tableDefinitionFilename,
                   const std::vector<int64_t>& subsystemConfigIdList);

    // *******************
    // *** Accessor(s) ***
    // *******************

    bool hasValidData() const {return _hasValidData;}

    std::vector<std::string> getSelectorColumnNames() const
    {
        return _configSelector.getSelectionColumns();
    }
    std::string getSelectorValueForDefault(std::string columnName) const
    {
        ConfigValueMap::const_iterator iter =
            _selectorDefaultValueMap.find(columnName);
        if (iter == _selectorDefaultValueMap.end()) {
            return NO_DEFAULT_KEY_STRING;
        }
        else {
            return iter->second;
        }
    }
    std::vector<std::string>
        getDistinctSelectorColumnValues(const std::string& columnName) const
    {
        ConfigValueListMap::const_iterator iter =
            _distinctSelectorValueMap.find(columnName);
        if (iter == _distinctSelectorValueMap.end()) {
            std::vector<std::string> emptyList;
            return emptyList;
        }
        else {
            return *(iter->second);
        }
    }
    std::vector<std::string> getDataColumnNames() const
    {
        return _dataColumnNameList;
    }
    std::vector<std::string> getDataColumnTypes() const
    {
        return _dataColumnTypeList;
    }
    std::string getTableName() const {return _tableName;}

    template<class T>
    bool getTrueValue(const std::string& dataColumnName, T& resultValue);

    template<class T, class U>
    bool getTrueValueFor1Key(const U& primarySelectionValue,
                             const std::string& dataColumnName,
                             T& resultValue);

    template<class T, class U, class V>
    bool getTrueValueFor2Keys(const U& primarySelectionValue,
                              const V& secondarySelectionValue,
                              const std::string& dataColumnName,
                              T& resultValue);

    template<class T, class U, class V, class W>
    bool getTrueValueFor3Keys(const U& primarySelectionValue,
                              const V& secondarySelectionValue,
                              const W& tertiarySelectionValue,
                              const std::string& dataColumnName,
                              T& resultValue);

    template<class T, class U>
    bool getEffectiveValueFor1Key(const U& primarySelectionValue,
                                  const std::string& dataColumnName,
                                  T& resultValue);

    template<class T, class U, class V>
    bool getEffectiveValueFor2Keys(const U& primarySelectionValue,
                                   const V& secondarySelectionValue,
                                   const std::string& dataColumnName,
                                   T& resultValue);

    template<class T, class U, class V, class W>
    bool getEffectiveValueFor3Keys(const U& primarySelectionValue,
                                   const V& secondarySelectionValue,
                                   const W& tertiarySelectionValue,
                                   const std::string& dataColumnName,
                                   T& resultValue);

    // *******************
    // *** Modifier(s) ***
    // *******************

    bool saveValues(const std::vector<dbutils::SimpleParameterSet>& newDataList) const;

 private:

    void _initialize(boost::shared_ptr<nova::database::Table> dbTablePtr,
                     const std::vector<int64_t>& subsystemConfigIdList,
                     const ConfigDataSelector& selector,
                     std::vector<std::string> dataColumnNameList,
                     int debugFlag);

    template<class T>
    bool _getValueFromContainer(const std::string& key,
                                const std::string& dataColumnName,
                                T& value);

    // ********************
    // *** Private data ***
    // ********************

    bool _hasValidData;
    std::string _tableDefinitionFilename;
    ConfigDataSelector _configSelector;
    ConfigValueMap _selectorDefaultValueMap;
    ConfigValueListMap _distinctSelectorValueMap;
    std::vector<std::string> _dataColumnNameList;
    std::vector<std::string> _dataColumnTypeList;
    std::string _tableName;
    ConfigDataContainer _resultSet;

    std::vector<std::string> _keyElementsForDefaults;
    std::vector<bool> _noKeyElementForDefaults;
};

/**
 * Fetches the requested value...
 */
template<class T>
bool ConfigDataTree::
getTrueValue(const std::string& dataColumnName, T& resultValue)
{
    std::string key = KEY_FOR_SYSTEM_WIDE_PARAM;
    return _getValueFromContainer(key, dataColumnName, resultValue);
}

/**
 * Fetches the requested value...
 */
template<class T, class U>
bool ConfigDataTree::
getTrueValueFor1Key(const U& primarySelectionValue,
                    const std::string& dataColumnName, T& resultValue)
{
    std::string key =
        boost::lexical_cast<std::string>(primarySelectionValue);
    return _getValueFromContainer(key, dataColumnName, resultValue);
}

/**
 * Fetches the requested value...
 */
template<class T, class U, class V>
bool ConfigDataTree::
getTrueValueFor2Keys(const U& primarySelectionValue,
                     const V& secondarySelectionValue,
                     const std::string& dataColumnName,
                     T& resultValue)
{
    std::string key =
        boost::lexical_cast<std::string>(primarySelectionValue) +
        KEY_SEPARATOR_STRING +
        boost::lexical_cast<std::string>(secondarySelectionValue);
    return _getValueFromContainer(key, dataColumnName, resultValue);
}

/**
 * Fetches the requested value...
 */
template<class T, class U, class V, class W>
bool ConfigDataTree::
getTrueValueFor3Keys(const U& primarySelectionValue,
                     const V& secondarySelectionValue,
                     const W& tertiarySelectionValue,
                     const std::string& dataColumnName,
                     T& resultValue)
{
    std::string key =
        boost::lexical_cast<std::string>(primarySelectionValue) +
        KEY_SEPARATOR_STRING +
        boost::lexical_cast<std::string>(secondarySelectionValue) +
        KEY_SEPARATOR_STRING +
        boost::lexical_cast<std::string>(tertiarySelectionValue);
    return _getValueFromContainer(key, dataColumnName, resultValue);
}

/**
 * Fetches the requested value...
 */
template<class T, class U>
bool ConfigDataTree::
getEffectiveValueFor1Key(const U& primarySelectionValue,
                         const std::string& dataColumnName, T& resultValue)
{
    std::string key =
        boost::lexical_cast<std::string>(primarySelectionValue);

    if (!_getValueFromContainer(key, dataColumnName, resultValue)) {
        if (_keyElementsForDefaults.size() < 1) {return false;}

        key = _keyElementsForDefaults[0];
        if (_noKeyElementForDefaults[0]) {
            boost::replace_all(key, NO_DEFAULT_KEY_STRING,
                               boost::lexical_cast<std::string>
                               (primarySelectionValue));
        }

        return _getValueFromContainer(key, dataColumnName, resultValue);
    }
    return true;
}

/**
 * Fetches the requested value...
 */
template<class T, class U, class V>
bool ConfigDataTree::
getEffectiveValueFor2Keys(const U& primarySelectionValue,
                          const V& secondarySelectionValue,
                          const std::string& dataColumnName,
                          T& resultValue)
{
    std::string key =
        boost::lexical_cast<std::string>(primarySelectionValue) +
        KEY_SEPARATOR_STRING +
        boost::lexical_cast<std::string>(secondarySelectionValue);

    if (! _getValueFromContainer(key, dataColumnName, resultValue)) {
        if (_keyElementsForDefaults.size() < 2) {return false;}

        std::string secondKeyElement = _keyElementsForDefaults[1];
        if (_noKeyElementForDefaults[1]) {
            boost::replace_all(secondKeyElement, NO_DEFAULT_KEY_STRING,
                               boost::lexical_cast<std::string>
                               (secondarySelectionValue));
        }
        key = boost::lexical_cast<std::string>(primarySelectionValue) +
            KEY_SEPARATOR_STRING + secondKeyElement;

        if (! _getValueFromContainer(key, dataColumnName, resultValue)) {
            std::string firstKeyElement = _keyElementsForDefaults[0];
            if (_noKeyElementForDefaults[0]) {
                boost::replace_all(firstKeyElement, NO_DEFAULT_KEY_STRING,
                                   boost::lexical_cast<std::string>
                                   (primarySelectionValue));
            }
            key = firstKeyElement + KEY_SEPARATOR_STRING + secondKeyElement;

            return _getValueFromContainer(key, dataColumnName,
                                          resultValue);
        }
    }
    return true;
}

/**
 * Fetches the requested value...
 */
template<class T, class U, class V, class W>
bool ConfigDataTree::
getEffectiveValueFor3Keys(const U& primarySelectionValue,
                          const V& secondarySelectionValue,
                          const W& tertiarySelectionValue,
                          const std::string& dataColumnName,
                          T& resultValue)
{
    std::string key =
        boost::lexical_cast<std::string>(primarySelectionValue) +
        KEY_SEPARATOR_STRING +
        boost::lexical_cast<std::string>(secondarySelectionValue) +
        KEY_SEPARATOR_STRING +
        boost::lexical_cast<std::string>(tertiarySelectionValue);

    if (! _getValueFromContainer(key, dataColumnName, resultValue)) {
        if (_keyElementsForDefaults.size() < 3) {return false;}

        std::string thirdKeyElement = _keyElementsForDefaults[2];
        if (_noKeyElementForDefaults[2]) {
            boost::replace_all(thirdKeyElement, NO_DEFAULT_KEY_STRING,
                               boost::lexical_cast<std::string>
                               (tertiarySelectionValue));
        }
        key = boost::lexical_cast<std::string>(primarySelectionValue) +
            KEY_SEPARATOR_STRING +
            boost::lexical_cast<std::string>(secondarySelectionValue) +
            KEY_SEPARATOR_STRING + thirdKeyElement;

        if (! _getValueFromContainer(key, dataColumnName, resultValue)) {
            std::string secondKeyElement = _keyElementsForDefaults[1];
            if (_noKeyElementForDefaults[1]) {
                boost::replace_all(secondKeyElement, NO_DEFAULT_KEY_STRING,
                                   boost::lexical_cast<std::string>
                                   (secondarySelectionValue));
            }
            key = boost::lexical_cast<std::string>(primarySelectionValue) +
                KEY_SEPARATOR_STRING + secondKeyElement +
                KEY_SEPARATOR_STRING + thirdKeyElement;

            if (! _getValueFromContainer(key, dataColumnName,
                                         resultValue)) {
                std::string firstKeyElement = _keyElementsForDefaults[0];
                if (_noKeyElementForDefaults[0]) {
                    boost::replace_all(firstKeyElement, NO_DEFAULT_KEY_STRING,
                                       boost::lexical_cast<std::string>
                                       (primarySelectionValue));
                }
                key = firstKeyElement +
                    KEY_SEPARATOR_STRING + secondKeyElement +
                    KEY_SEPARATOR_STRING + thirdKeyElement;

                return _getValueFromContainer(key, dataColumnName,
                                              resultValue);
            }
        }
    }
    return true;
}

/**
 * 
 */
template<class T>
bool ConfigDataTree::
_getValueFromContainer(const std::string& key,
                       const std::string& dataColumnName,
                       T& value)
{
    ConfigDataContainer::const_iterator iter1 = _resultSet.find(key);
    if (iter1 == _resultSet.end()) {return false;}

    ConfigValueMapPtr mapPtr = _resultSet[key];
    ConfigValueMap::const_iterator iter2 = mapPtr->find(dataColumnName);
    if (iter2 == mapPtr->end()) {return false;}

    value = boost::lexical_cast<T>(iter2->second);
    return true;
}

};  // end of namespace daqconfig
};  // end of namespace dbutils

#endif
