#ifndef _RMS_CONSUMER_H
#define _RMS_CONSUMER_H

/**
 * @package gov.fnal.cd.rms
 *
 * @brief Classes that allow someone to send and receive
 * messages utilizing the Responsive Messaging System.
 *
 * This namespace includes the Producer and Consumer classes
 * as well as interfaces that can be implemented to handle
 * messages straight out of RMS.
 */

#include <rms/MessageFilter.h>
#include <rms/ClientListenerLoop.h>
#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessage.h>
#include <rms/base/RmsStatus.h>
#include <rms/provider/BufferedProviderListener.h>
#include <rms/provider/RmsConnection.h>
#include <rms/util/LinkedBlockingQueue.h>

#include <boost/thread/thread.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/bind.hpp>

#include <string>
#include <iostream>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

/**
 * This class contains functionality for receiving messages and sending
 * replies.
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.2 $ $Date: 2010/04/16 18:40:37 $
 */
class RmsConsumer : public MessageFilter {
 public:
    RmsConsumer(boost::shared_ptr<provider::RmsConnection> providerConnection, 
                base::RmsDestination dest);
    RmsConsumer(boost::shared_ptr<provider::RmsConnection> providerConnection,
                std::vector <base::RmsDestination> dest);
    ~RmsConsumer();

    boost::shared_ptr<std::string> receiveString();
    boost::shared_ptr<std::string> receiveString(long timeout);
    boost::shared_ptr<std::string> receiveStringNoWait();

    boost::shared_ptr<base::RmsMessage> receiveMessage();
    boost::shared_ptr<base::RmsMessage> receiveMessage(long timeout);
    boost::shared_ptr<base::RmsMessage> receiveMessageNoWait();
    
    void setMessageListener(RmsMessageListener *listener);
    
    void sendReply(boost::shared_ptr<base::RmsMessage> message);
    
    void close();
    bool verify(boost::shared_ptr<base::RmsMessage> message);
    
    void handleLoopback(boost::shared_ptr<base::RmsMessage> loopbackMessage);

 private:
    /**
     * Handle to the RMS provider
     */
    boost::shared_ptr<provider::RmsConnection> _providerConnection; 

    /** 
     * RmsDestination that we are listening to 
     */
    std::vector <base::RmsDestination> _myDestinations;
    
    /**
     * The client listener loop that pulls data off of the input queue and
     * calls our listener 
     */
    ClientListenerLoop _listenerLoop;
       
    /**
     * Listens for messages and places them onto the input queue.
     */
    provider::BufferedProviderListener *_providerListener;

    /**
     * The input queue that messages are put onto by the provider, pulled 
     * off of by the client listener loop. 
     */
    util::LinkedBlockingQueue <boost::shared_ptr<std::string> > _inputQueue;

    /**
     * Handle to the listenerThread that is created in the constructor.
     */
    boost::thread _listenerThread;
    
    /** 
     * RmsDestinations that we use for verifying messages
     */
    std::vector <base::RmsDestination> _myVerifyDests;
    
};

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
