#ifndef _RMS_LOCALHOSTCONNECTION_H
#define _RMS_LOCALHOSTCONNECTION_H

#include <rms/base/RmsDestination.h>
#include <rms/base/RmsMessage.h>
#include <rms/provider/RmsConnection.h>

#include <boost/shared_ptr.hpp>

#include <string>
#include <vector>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace provider {

/**
 * A simple provider useful for testing the RMS interface
 * routines.  This provider will route messages between
 * consumers and producers that are connected to it.  
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */

class LocalhostConnection : public RmsConnection {
 public:
    LocalhostConnection (const std::string applicationName, 
                         const int applicationPartition);
    
    bool supportsDestination(const base::RmsDestination& candidateDestination);
    
    void sendString(const base::RmsDestination& dest, const std::string& messageString);
    void sendMessage(const base::RmsDestination& dest, const base::RmsMessage& message);
    
    void addListener(const base::RmsDestination& dest, ProviderListener *listener);
    void removeListener(const base::RmsDestination& dest, ProviderListener *listener);
    
    bool ping(const base::RmsDestination& pingDestination);
    
    void close();

 private:
    /**
     * Container for a destination and a provider listener. The
     * destination and the provider listener are placed into this
     * object before being put into the destinationList for processing.
     */
    class ListenerContainer {
    public:
        /**
         * Create an empty listener container.
         */
        ListenerContainer() { 
            return; 
        }
    
        /**
         * Create a listener container given a destination object and
         * a reference to a listener object.
         *
         * @param dest The destination object
         * @param listener A reference to a listener object
         */
        ListenerContainer(base::RmsDestination dest, 
                          ProviderListener *listener) {
            _dest = dest;
            _listener = listener;
        }
        
        /**
         * Get the destination object that is associated with this
         * listener container.
         *
         * @return The destination object
         */
        base::RmsDestination getDestination() {
            return _dest;
        }
        
        /**
         * Get the reference to the listener object that is associated
         * with this listener container.
         *
         * @return A reference to the listener object
         */
        ProviderListener *getListener() {
            return _listener;
        }
        
        /**
         * Set the destination object in this container.
         *
         * @param dest The destination object to copy into
         * this container.
         */
        void setDestination(base::RmsDestination dest) {
            _dest = dest;
        }
        
        /**
         * Set the listener object in this container.
         *
         * @param listener A reference to the listener object
         * this is to be called when a message is received that
         * matches the destination associated with this 
         * listener.
         */
        void setListener(ProviderListener *listener) {
            _listener = listener;
        }
    private:
        /**
         * The destination for this listener.  As messages are received 
         * they will be compared against this destination to determine 
         * if they should be forwarded to this listener. 
         */
        base::RmsDestination _dest;

        /**
         * The listener to send messages to if they match the destination. 
         */
        ProviderListener *_listener; 

    };

    /**
     * A list of destinations and listeners that have been associated 
     * with this provider. 
     */
    std::vector <ListenerContainer> _destinationList; 
};

};  // end of namespace provider

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
