#ifndef _RMS_MESSAGEFRAGMENT_H
#define _RMS_MESSAGEFRAGMENT_H

#include <string>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace provider {

/**
 * Message fragment class.  This class is used to facilitate the
 * transmission of string messages that need to be broken
 * into multiple parts because of some size constraint in the system.
 * It does not enforce any size constraints itself, but instead relies
 * on users to only create payloads of appropriate sizes.
 *
 * There are currently two headers defined, version 0:
 * - fragment prefix = "RMS" -> 3 bytes
 * - header version -> 1 byte
 * - fragment index -> 2 bytes (zero-based)
 * - fragment count -> 2 bytes (one-based)
 * - payload size -> 4 bytes
 * - message UUID -> 16 bytes
 *
 * Version 1:
 * - fragment prefix = "RMS" -> 3 bytes
 * - header version -> 1 byte
 * - fragment index -> 2 bytes (zero-based)
 * - fragment count -> 2 bytes (one-based)
 * - payload size -> 4 bytes
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */

class MessageFragment {
 public:
    MessageFragment(const char *guid, int index, int totalCount);
    MessageFragment(char *fragment);
    
    ~MessageFragment();

    int getFragmentIndex() const;
    int getFragmentCount() const;
    char *getMessageId() const;
    int getPayloadSize() const;
    
    void setPayload(std::string payloadString);
    std::string getPayload() const;
    void getBytes(char **message, int *size) const;
    void clearPayload();
    
    static int getMaxHeaderSize();

 private:

    /**
     * Structure representing the version 0 fragment header,
     * which is documented in further detail above.
     */
    struct fragmentHeader {
        char fragmentPrefix[3];
        char headerVersion;
        unsigned short fragmentIndex;
        unsigned short fragmentCount;
        unsigned int payloadSize;
        char uuid[16];
    };
    
    /**
     * The prefix that all message fragments have.
     */
    static const std::string _FRAGMENT_PREFIX;
    
    /**
     * A string consisting of the message fragment's payload.
     */
    std::string _messageFragmentPayload;
    
    /**
     * The index of this fragment with respect to the 
     * entire message.
     */
    int _fragmentIndex;

    /**
     * The total number of fragments that make up the message.
     */
    int _fragmentCount;

    /**
     * The UUID of this message.
     */
    char *_uuid;
};

};  // end of namespace provider

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
