#ifndef _RMS_TIMEUTILS_H
#define _RMS_TIMEUTILS_H

#include <string>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/algorithm/string/replace.hpp>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace util {

/**
 * Utilities for dates and times.
 *
 * The standard format for RMS datetimes is "yyyy-MM-dd HH:mm:ss.SSS'Z'".
 *
 * @author Kurt Biery
 * @author Steve Foulkes
 * @version $Revision: 1.1.1.1 $ $Date: 2009/10/09 20:03:00 $
 */

class TimeUtils {
 public:    
    /**
     * Returns a Boost Ptime object initialized to universal time (GMT).
     *
     * @return The current time in GMT.
     */
    static boost::posix_time::ptime getUniversalPtime() {
        boost::posix_time::ptime 
            now(boost::posix_time::microsec_clock::universal_time());
        return now;
    }
    
    /**
     * Converts the specified Boost Ptime to a datetime string in the
     * standard RMS date format.
     *
     * @param inputTime The time to convert to the RMS date format.
     *
     * @return A string representing the time in inputTime.
     */
    static std::string 
        convertPtimeToString(boost::posix_time::ptime inputTime) {
        
        // determine the GMT offset (probably need to do only once...)
        boost::posix_time::time_duration timeDiff = 
            boost::posix_time::second_clock::local_time() -
            boost::posix_time::second_clock::universal_time();
        boost::posix_time::time_duration gmtOffset = 
            boost::posix_time::hours(timeDiff.hours());
        
        // convert the input time to local time (ideally we shouldn't 
        // have to do this, but the Java date conversion currently wants 
        // local time with the GMT offset appended to the string)
        inputTime += gmtOffset;

        // create the string
        std::string convertedString = to_iso_extended_string(inputTime);

        // strip off any extra fractional seconds digits (more than three)
        size_t dotPos = convertedString.find_last_of(".");
        if (dotPos != std::string::npos &&
            convertedString.length() > (dotPos + 4)) {
            convertedString = convertedString.substr(0, (dotPos + 4));
        }

        // append the TZ offset
        std::string tzOffset = to_simple_string(gmtOffset);
        size_t pos = tzOffset.find(":");
        if (pos != std::string::npos) {
            pos = tzOffset.find(":", pos+1);
            if (pos != std::string::npos) {
                tzOffset = tzOffset.substr(0, pos);
            }
        }
        if (gmtOffset.hours() >= 0 &&
            tzOffset.find("+") == std::string::npos) {
            tzOffset.insert(0, "+");
        }
        convertedString.append(tzOffset);

        // return the datetime string
        return convertedString;
    }

    /**
     * Converts the specified datetime string (in the standard RMS date format)
     * to a Boost Ptime.
     *
     * @param inputString A string with the time in the RMS format.
     *
     * @return A boost ptime object initialized to the time in inputString.
     */
    static boost::posix_time::ptime 
        convertStringToPtime(std::string inputString) {
        // determine the time zone offset
        boost::posix_time::time_duration timeOffset;
        size_t timePos = inputString.find("T");
        if (timePos != std::string::npos) {
            size_t offsetPos = inputString.find("-", timePos);
            if (offsetPos == std::string::npos) {
                offsetPos = inputString.find("+", timePos);
            }

            if (offsetPos != std::string::npos) {
                timeOffset = boost::posix_time::duration_from_string(inputString.substr(offsetPos));
                inputString = inputString.substr(0, offsetPos);
            }
        }

        // convert the string to a time
        boost::replace_all(inputString, "T", " ");
        boost::posix_time::ptime 
            convertedTime(boost::posix_time::time_from_string(inputString));

        // fold in the TZ offset (to get GMT)
        convertedTime = convertedTime - timeOffset;
        
        // return the time
        return convertedTime;
    }

};

};  // end of namespace util

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov

#endif
