///
/// \file    HardwareDisplay.cpp
/// \brief   Map channel addresses to a sensible (x,y) coordinate
/// \author  messier@indiana.edu
/// \version $Id: HardwareDisplay.cpp,v 1.3 2014/06/17 19:28:53 messier Exp $
///
#include "DAQChannelMap/HardwareDisplay.h"
#include "DAQChannelMap/DAQChannelMap.h"
#include <iostream>
#include <stdlib.h>
#include "NovaDAQConventions/DAQConventions.h"

using namespace daqchannelmap;
using namespace novadaq::cnv;

void HardwareDisplay::SetupDet(int det) {
  switch (det) {
  case kNDOS:    this->SetupNDOS();    break;
  case kNEARDET: this->SetupNearDet(); break;
  case kFARDET:  this->SetupFarDet();  break;
  default:
    std::cerr << __FILE__ << " : " << __LINE__
	      <<    " Invalid detector type " << det
	      << std::endl;
    abort();
  }
}

//......................................................................

void HardwareDisplay::SetupNDOS() 
{
  fDetId = kNDOS;

  fNdiblocks = 4;
  fNdcmTop   = 1;
  fNdcmSide  = 2;
  fNfebCols  = 16;
  fNfebRows  = 4;
  fNpixCols  = 4;
  fNpixRows  = 8;

  fNxPerDB  = fNfebCols*fNpixCols;
  fNyPerDCM = fNfebRows*fNpixRows;

  fXmax = fNdiblocks*fNxPerDB;
  fYmax = (fNdcmTop+fNdcmSide)*fNyPerDCM;

  fNplanesPerBL = 31;
  fNplanes      = 2*fNplanesPerBL*fNdiblocks; // 31 planes per block
  fNxcells      = 32*2;                 // 32 cells / module, 2 modules
  fNycells      = 32*3;                 // 32 cells / module, 3 modules
  fNcellPerMod  = 32;
}

//......................................................................

void HardwareDisplay::SetupNearDet() 
{
  fDetId = kNEARDET;

  fNdiblocks = 4;
  fNdcmTop   = 2;
  fNdcmSide  = 2;
  fNfebCols  = 16;
  fNfebRows  = 4;
  fNpixCols  = 4;
  fNpixRows  = 8;

  fNxPerDB  = fNfebCols*fNpixCols;
  fNyPerDCM = fNfebRows*fNpixRows;

  fXmax = fNdiblocks*fNxPerDB;
  fYmax = (fNdcmTop+fNdcmSide)*fNyPerDCM;

  fNplanesPerBL = 32;
  fNplanes      = 2*fNplanesPerBL*fNdiblocks; // 32 planes per block
  fNxcells      = 32*3;                 // 32 cells / module, 3 modules
  fNycells      = 32*3;                 // 32 cells / module, 3 modules
  fNcellPerMod  = 32;
}

//......................................................................

void HardwareDisplay::SetupFarDet() 
{
  fDetId = kFARDET;

  fNdiblocks = 14;
  fNdcmTop   = 6;
  fNdcmSide  = 6;
  fNfebCols  = 16;
  fNfebRows  = 4;
  fNpixCols  = 4;
  fNpixRows  = 8;

  fNxPerDB  = fNfebCols*fNpixCols;
  fNyPerDCM = fNfebRows*fNpixRows;

  fXmax = fNdiblocks*fNxPerDB;
  fYmax = (fNdcmTop+fNdcmSide)*fNyPerDCM;

  fNplanesPerBL = 32;
  fNplanes      = 2*fNplanesPerBL*fNdiblocks; // 32 planes per block
  fNxcells      = 32*12;                // 32 cells / module, 12 modules
  fNycells      = 32*12;                // 32 cells / module, 12 modules
  fNcellPerMod  = 32;
}

//......................................................................

void HardwareDisplay::DiblockXY(unsigned int db, 
				unsigned int *ix,
				unsigned int *iy)
{
  //*ix = fNxPerDB*(db-1);   // diblocks numbered L to R
  *ix = fNxPerDB*(fNdiblocks - db);   // diblocks numbered R to L
  *iy = 0;
}

//......................................................................

void HardwareDisplay::DCMXY(unsigned int dcm,
			    unsigned int* ix,
			    unsigned int* iy)
{
  *ix = 0;
  *iy = fYmax - dcm*fNyPerDCM;
}

//......................................................................

void HardwareDisplay::FEBXY(unsigned int feb,
			    unsigned int* ix,
			    unsigned int* iy)
{
  *ix = (feb%fNfebCols)*fNpixCols;
  *iy = ((fNfebRows - 1) - feb/fNfebCols)*fNpixRows;
}

//......................................................................

void HardwareDisplay::PixXY(unsigned int pix,
			    unsigned int* ix,
			    unsigned int* iy)
{
  static unsigned int x[32] = {
    3,3,3,3,3,3,3,3, 2,2,2,2,2,2,2,2,
    1,1,1,1,1,1,1,1, 0,0,0,0,0,0,0,0
  };
  static unsigned int y[32] = {
    1,0,3,2,5,4,7,6,  1,0,3,2,5,4,6,7,
    7,6,4,5,2,3,0,1,  6,7,4,5,2,3,0,1
  };

  *ix = x[pix];
  *iy = y[pix];
}

//......................................................................

void HardwareDisplay::DAQToXY(unsigned int db,
			      unsigned int dcm,
			      unsigned int feb,
			      unsigned int pix,
			      unsigned int* ix,
			      unsigned int* iy) 
{
  unsigned int x1, x2, x3, x4;
  unsigned int y1, y2, y3, y4;
  
  this->DiblockXY(db,  &x1, &y1);
  this->DCMXY    (dcm, &x2, &y2);
  this->FEBXY    (feb, &x3, &y3);
  this->PixXY    (pix, &x4, &y4);

  *ix = x1+x2+x3+x4;
  *iy = y1+y2+y3+y4;
}





//......................................................................

void HardwareDisplay::XYDiblock(unsigned int *db, 
				unsigned int ix)
{
  //*db = 1 + ix/fNxPerDB;   // diblocks numbered L to R
  *db = fNdiblocks - ix/fNxPerDB;   //diblocks numbered R to L
}

//......................................................................

void HardwareDisplay::XYDCM(unsigned int* dcm,
			    unsigned int iy)
{
  *dcm = (fNdcmTop + fNdcmSide) - iy/fNyPerDCM;
}

//......................................................................

void HardwareDisplay::XYFEB(unsigned int* feb,
			    unsigned int ix,
			    unsigned int iy)
{
  unsigned int x1 = (ix%fNxPerDB)/fNpixCols;
  unsigned int y1 = (iy%fNyPerDCM)/fNpixRows;

  *feb = fNfebCols*(fNfebRows - 1 - y1) + x1;
}

//......................................................................

void HardwareDisplay::XYPix(unsigned int* pix,
			    unsigned int ix,
			    unsigned int iy)
{
  static unsigned int P[8][4] = {{30,22,9,1},
				 {31,23,8,0},
				 {28,20,11,3},
				 {29,21,10,2},
				 {26,18,13,5},
				 {27,19,12,4},
				 {24,17,14,7},
				 {25,16,15,6}};

  unsigned int x1 = ix%fNpixCols;
  unsigned int y1 = iy%fNpixRows;

  *pix = P[y1][x1];
}

//......................................................................

void HardwareDisplay::XYToDAQ(unsigned int* db,
			      unsigned int* dcm,
			      unsigned int* feb,
			      unsigned int* pix,
			      unsigned int ix,
			      unsigned int iy) 
{
  this->XYDiblock(db,  ix);
  this->XYDCM    (dcm, iy);
  this->XYFEB    (feb, ix, iy);
  this->XYPix    (pix, ix, iy);
}

//......................................................................

void HardwareDisplay::DAQToPlaneModCell(unsigned int* block,
					unsigned int* plane,
					unsigned int* view,
					unsigned int* module,
					unsigned int* cell,
					unsigned int  db,
					unsigned int  dcm,
					unsigned int  feb,
					unsigned int  pix)
{
  daqchannelmap::DAQChannelMap*
    cmap = daqchannelmap::DAQChannelMap::getInstance(fDetId);
  
  unsigned int dchan = cmap->encodeDChan(fDetId, db, dcm, feb, pix);
  unsigned int lchan = cmap->encodeLChan(dchan);
  
  *block  = cmap->computeBlock(lchan);
  *plane  = cmap->getPlane(lchan);
  *cell   = cmap->getCell(lchan);
  *view   = cmap->getView(dchan);
  *module = ((*cell)/fNcellPerMod);
}

//......................................................................

void HardwareDisplay::DAQToPlaneInBlockModCell(unsigned int* block,
					       unsigned int* pinb,
					       unsigned int* view,
					       unsigned int* module,
					       unsigned int* cell,
					       unsigned int  db,
					       unsigned int  dcm,
					       unsigned int  feb,
					       unsigned int  pix)
{
  daqchannelmap::DAQChannelMap*
    cmap = daqchannelmap::DAQChannelMap::getInstance(fDetId);
  
  unsigned int dchan = cmap->encodeDChan(fDetId, db, dcm, feb, pix);
  unsigned int lchan = cmap->encodeLChan(dchan);
  
  *block  = cmap->computeBlock(lchan);
  *pinb   = cmap->computePlaneInBlock(lchan);
  *cell   = cmap->getCell(lchan);
  *view   = cmap->getView(dchan);
  *module = ((*cell)/fNcellPerMod);
}

//......................................................................

void HardwareDisplay::PlaneCellToDAQ(unsigned int  plane,
				     unsigned int* pinb,
				     unsigned int* view,
				     unsigned int* module,
				     unsigned int  cell,
				     unsigned int* db,
				     unsigned int* block,
				     unsigned int* dcm,
				     unsigned int* feb,
				     unsigned int* pix)
{
  daqchannelmap::DAQChannelMap*
    cmap = daqchannelmap::DAQChannelMap::getInstance(fDetId);
  
  unsigned int lchan = cmap->encodeLChan(fDetId, plane, cell);
  unsigned int dchan = cmap->encodeDChan(lchan);
  
  *pinb   = cmap->computePlaneInBlock(lchan);
  *view   = cmap->getView(dchan);
  *db     = cmap->getDiBlock(dchan);
  *block  = cmap->computeBlock(lchan);
  *dcm    = cmap->getDCM    (dchan);
  *feb    = cmap->getFEB(dchan);
  *pix    = cmap->getPixel  (dchan);
  *module = ((cell)/fNcellPerMod);
}

///////////////////////////////////////////////////////////////////////
