#include <DatabaseUtils/DAQConfig/ConfigDataTree.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/DBColumnUtils.h>
#include <DatabaseUtils/DAQConfig/GlobalNameSpec.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <NovaDatabase/Table.h>
#include <boost/lexical_cast.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/algorithm/string.hpp>

namespace NOVADB = nova::database;
using namespace dbutils::daqconfig;
using namespace nova::database;

class GlobalCleanupHandler
{
  public:

    bool noUpdate;
    int64_t globalConfigId;
    std::string daqOrSlowControls;
    bool success;

    GlobalCleanupHandler() :
        noUpdate(true), globalConfigId(-1),success(false) {}

    ~GlobalCleanupHandler()
    {
        if (noUpdate) {return;}

        if (globalConfigId > 0) {
            IDSpec idSpec(globalConfigId, IDSpec::GLOBAL,
                          (daqOrSlowControls == "DAQ" ?
                           IDSpec::DAQ : IDSpec::SLOWCONTROLS));
            if (success) {
                if (! NamedConfigUtils::markConfigDone(idSpec, true)) {
                    std::cout << "ERROR: unable to mark the new global "
                              << "configuration complete in the database."
                              << std::endl;
                }
            }
            else {
                if (! NamedConfigUtils::markConfigDone(idSpec, false)) {
                    std::cout << "WARNING: unable to modify the database to "
                              << "show that the new global is done changing."
                              << std::endl;
                }
            }
        }
    }
};

class SubsystemCleanupHandler
{
  public:

    bool noUpdate;
    int64_t subsystemConfigId;
    std::string daqOrSlowControls;
    bool success;

    SubsystemCleanupHandler() :
        noUpdate(true), subsystemConfigId(-1), success(false) {}

    ~SubsystemCleanupHandler()
    {
        if (noUpdate) {return;}

        if (subsystemConfigId > 0) {
            IDSpec idSpec(subsystemConfigId, IDSpec::SUBSYSTEM,
                          (daqOrSlowControls == "DAQ" ?
                           IDSpec::DAQ : IDSpec::SLOWCONTROLS));
            if (success) {
                if (! NamedConfigUtils::markConfigDone(idSpec, true)) {
                    std::cout << "ERROR: unable to mark the new subsystem "
                              << "configuration complete in the database."
                              << std::endl;
                }
            }
            else {
                if (! NamedConfigUtils::markConfigDone(idSpec, false)) {
                    std::cout << "WARNING: unable to modify the database to "
                              << "show that the new subsystem is done changing."
                              << std::endl;
                }
            }
        }
    }
};

bool _populateTableByDCMFEBAndThirdSelector(bool doNoUpdates,
                                            boost::shared_ptr<NOVADB::Table> dbt,
                                            const int64_t& subsysCfgId,
                                            ConfigDataTree& dataTree,
                                            const std::string& thirdSelectorName,
                                            const std::vector<std::string>&
                                            dataColumnNameList,
                                            const std::vector<std::string>&
                                            dcmNameList)
{
    std::vector<std::string> thirdSelectorValues =
        dataTree.getDistinctSelectorColumnValues(thirdSelectorName);

    int changeCount = 0;
    for (int32_t pvIdx = 0; pvIdx < (int) thirdSelectorValues.size(); ++pvIdx) {
        std::string thirdSelectorValue = thirdSelectorValues[pvIdx];

        for (uint32_t dcmIdx = 0; dcmIdx < dcmNameList.size(); ++dcmIdx) {
            std::string dcmName = dcmNameList[dcmIdx];
            for (int32_t febIdx = -1; febIdx < 64; ++febIdx) {

                bool aValueWasFound = false;
                boost::shared_ptr<NOVADB::Row> newRow;
                for (uint32_t colIdx = 0; colIdx < dataColumnNameList.size();
                     ++colIdx) {
                    std::string colName = dataColumnNameList[colIdx];
                    std::string value;
                    if (dataTree.getTrueValueFor3Keys(thirdSelectorValue,
                                                      dcmName, febIdx,
                                                      colName, value)) {
                        if (! doNoUpdates) {
                            if (newRow.get() == 0) {
                                newRow.reset(dbt->NewRow());
                                if (newRow.get() == 0) {return false;}
                            }
                            newRow->Set(colName, value);
                        }
                        aValueWasFound = true;
                    }
                }
                if (newRow.get() != 0) {
                    newRow->Set(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                                subsysCfgId);
                    newRow->Set(DBColumnParams::DCMID_COLUMN_NAME, dcmName);
                    newRow->Set(DBColumnParams::FEBID_COLUMN_NAME, febIdx);
                    newRow->Set(thirdSelectorName, thirdSelectorValue);
                    dbt->AddRow(newRow.get());
                }
                if (aValueWasFound) {++changeCount;}
            }
        }
    }
    std::cout << "Number of rows to be copied = " << changeCount
              << std::endl;
    return true;
}


bool _populateTableByDCMFEBAndPixel(bool doNoUpdates,
                                    boost::shared_ptr<NOVADB::Table> dbt,
                                    const int64_t& subsysCfgId,
                                    ConfigDataTree& dataTree,
                                    const std::vector<std::string>&
                                    dataColumnNameList,
                                    const std::vector<std::string>&
                                    dcmNameList)
{
    int changeCount = 0;
    for (uint32_t dcmIdx = 0; dcmIdx < dcmNameList.size(); ++dcmIdx) {
        std::string dcmName = dcmNameList[dcmIdx];
        for (int32_t febIdx = -1; febIdx < 64; ++febIdx) {
            for (int32_t pixelIdx = -1; pixelIdx < 64; ++pixelIdx) {

                bool aValueWasFound = false;
                boost::shared_ptr<NOVADB::Row> newRow;
                for (uint32_t colIdx = 0; colIdx < dataColumnNameList.size();
                     ++colIdx) {
                    std::string colName = dataColumnNameList[colIdx];
                    std::string value;
                    if (dataTree.getTrueValueFor3Keys(dcmName, febIdx,
                                                      pixelIdx, colName,
                                                      value)) {
                        //std::cout << colName << " = " << value
                        //          << " (dcm=" << dcmName
                        //          << ", feb=" << febIdx
                        //          << ", pixel=" << pixelIdx
                        //          << ")" << std::endl;

                        if (! doNoUpdates) {
                            if (newRow.get() == 0) {
                                newRow.reset(dbt->NewRow());
                                if (newRow.get() == 0) {return false;}
                            }
                            newRow->Set(colName, value);
                        }
                        aValueWasFound = true;
                    }
                }
                if (newRow.get() != 0) {
                    newRow->Set(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                                subsysCfgId);
                    newRow->Set(DBColumnParams::DCMID_COLUMN_NAME, dcmName);
                    newRow->Set(DBColumnParams::FEBID_COLUMN_NAME, febIdx);
                    newRow->Set(DBColumnParams::PIXELID_COLUMN_NAME, pixelIdx);
                    dbt->AddRow(newRow.get());
                }
                if (aValueWasFound) {++changeCount;}
            }
        }
    }
    std::cout << "Number of rows to be copied = " << changeCount
              << std::endl;
    return true;
}

bool _populateTableByDCMAndFEB(bool doNoUpdates,
                               boost::shared_ptr<NOVADB::Table> dbt,
                               const int64_t& subsysCfgId,
                               ConfigDataTree& dataTree,
                               const std::vector<std::string>&
                               dataColumnNameList,
                               const std::vector<std::string>& dcmNameList)
{
    int changeCount = 0;
    for (uint32_t dcmIdx = 0; dcmIdx < dcmNameList.size(); ++dcmIdx) {
        std::string dcmName = dcmNameList[dcmIdx];
        for (int32_t febIdx = -1; febIdx < 64; ++febIdx) {

            bool aValueWasFound = false;
            boost::shared_ptr<NOVADB::Row> newRow;
            for (uint32_t colIdx = 0; colIdx < dataColumnNameList.size();
                 ++colIdx) {
                std::string colName = dataColumnNameList[colIdx];
                std::string value;
                if (dataTree.getTrueValueFor2Keys(dcmName, febIdx,
                                                  colName,
                                                  value)) {
                    //std::cout << colName << " = " << value
                    //          << " (dcm=" << dcmName
                    //          << ", feb=" << febIdx
                    //          << ")" << std::endl;

                    if (! doNoUpdates) {
                        if (newRow.get() == 0) {
                            newRow.reset(dbt->NewRow());
                            if (newRow.get() == 0) {return false;}
                        }
                        newRow->Set(colName, value);
                    }
                    aValueWasFound = true;
                }
            }
            if (newRow.get() != 0) {
                newRow->Set(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                            subsysCfgId);
                newRow->Set(DBColumnParams::DCMID_COLUMN_NAME, dcmName);
                newRow->Set(DBColumnParams::FEBID_COLUMN_NAME, febIdx);
                dbt->AddRow(newRow.get());
            }
            if (aValueWasFound) {++changeCount;}
        }
    }
    std::cout << "Number of rows to be copied = " << changeCount
              << std::endl;
    return true;
}

bool _populateTableByTDUAndDelayType(bool doNoUpdates,
                               boost::shared_ptr<NOVADB::Table> dbt,
                               const int64_t& subsysCfgId,
                               ConfigDataTree& dataTree,
                               const std::vector<std::string>&
                               dataColumnNameList,
                               const std::vector<std::string>& tduNameList)
{
    int changeCount = 0;
    std::string delayTypeList[] = {"early","total","side","top"};;
    for (uint32_t tduIdx = 0; tduIdx < tduNameList.size(); ++tduIdx) {
        std::string tduName = tduNameList[tduIdx];
        for (unsigned int delayType = 0; delayType < 4; ++delayType) {

            bool aValueWasFound = false;
            boost::shared_ptr<NOVADB::Row> newRow;
            std::string delayTypeName = delayTypeList[delayType];
            for (uint32_t colIdx = 0; colIdx < dataColumnNameList.size();
                 ++colIdx) {
                std::string colName = dataColumnNameList[colIdx];
                std::string value;
                if (dataTree.getTrueValueFor2Keys(tduName, delayTypeName,
                                                  colName,
                                                  value)) {
                    //std::cout << colName << " = " << value
                    //          << " (tdu=" << tduName
                    //          << ", feb=" << febIdx
                    //          << ")" << std::endl;

                    if (! doNoUpdates) {
                        if (newRow.get() == 0) {
                            newRow.reset(dbt->NewRow());
                            if (newRow.get() == 0) {return false;}
                        }
                        newRow->Set(colName, value);
                    }
                    aValueWasFound = true;
                }
            }
            if (newRow.get() != 0) {
                newRow->Set(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                            subsysCfgId);
                newRow->Set(DBColumnParams::DCMID_COLUMN_NAME, tduName);
                newRow->Set(DBColumnParams::TDUDELAYTYPE_COLUMN_NAME, delayTypeName);
                dbt->AddRow(newRow.get());
            }
            if (aValueWasFound) {++changeCount;}
        }
    }
    std::cout << "Number of rows to be copied = " << changeCount
              << std::endl;
    return true;
}

bool _populateTableByDCM(bool doNoUpdates,
                         boost::shared_ptr<NOVADB::Table> dbt,
                         const int64_t& subsysCfgId,
                         ConfigDataTree& dataTree,
                         const std::vector<std::string>& dataColumnNameList,
                         const std::vector<std::string>& dcmNameList)
{
    for (uint32_t dcmIdx = 0; dcmIdx < dcmNameList.size(); ++dcmIdx) {
        std::string dcmName = dcmNameList[dcmIdx];

        boost::shared_ptr<NOVADB::Row> newRow;
        for (uint32_t colIdx = 0; colIdx < dataColumnNameList.size();
             ++colIdx) {
            std::string colName = dataColumnNameList[colIdx];
            std::string value;
            if (dataTree.getTrueValueFor1Key(dcmName, colName,
                                             value)) {
                std::cout << colName << " = " << value
                          << " (dcm=" << dcmName
                          << ")" << std::endl;

                if (! doNoUpdates) {
                    if (newRow.get() == 0) {
                        newRow.reset(dbt->NewRow());
                        if (newRow.get() == 0) {return false;}
                    }
                    newRow->Set(colName, value);
                }
            }
        }
        if (newRow.get() != 0) {
            newRow->Set(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                        subsysCfgId);
            newRow->Set(DBColumnParams::DCMID_COLUMN_NAME, dcmName);
            dbt->AddRow(newRow.get());
        }
    }
    return true;
}

bool _populateTableByHost(bool doNoUpdates,
                          boost::shared_ptr<NOVADB::Table> dbt,
                          const int64_t& subsysCfgId,
                          ConfigDataTree& dataTree,
                          const std::vector<std::string>& dataColumnNameList,
                          const std::vector<std::string>& hostNameList)
{
    for (uint32_t hostIdx = 0; hostIdx < hostNameList.size(); ++hostIdx) {
        std::string hostName = hostNameList[hostIdx];

        boost::shared_ptr<NOVADB::Row> newRow;
        for (uint32_t colIdx = 0; colIdx < dataColumnNameList.size();
             ++colIdx) {
            std::string colName = dataColumnNameList[colIdx];
            std::string value;
            if (dataTree.getTrueValueFor1Key(hostName, colName,
                                             value)) {
                std::cout << colName << " = " << value
                          << " (host=" << hostName
                          << ")" << std::endl;

                if (! doNoUpdates) {
                    if (newRow.get() == 0) {
                        newRow.reset(dbt->NewRow());
                        if (newRow.get() == 0) {return false;}
                    }
                    newRow->Set(colName, value);
                }
            }
        }
        if (newRow.get() != 0) {
            newRow->Set(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                        subsysCfgId);
            newRow->Set(DBColumnParams::HOSTNAME_COLUMN_NAME, hostName);
            dbt->AddRow(newRow.get());
        }
    }
    return true;
}

bool _populateTable(bool doNoUpdates,
                    boost::shared_ptr<NOVADB::Table> dbt,
                    const int64_t& subsysCfgId,
                    ConfigDataTree& dataTree,
                    const std::vector<std::string>& dataColumnNameList)
{
    boost::shared_ptr<NOVADB::Row> newRow;
    for (uint32_t colIdx = 0; colIdx < dataColumnNameList.size(); ++colIdx) {
        std::string colName = dataColumnNameList[colIdx];
        std::string value;
        if (dataTree.getTrueValue(colName, value)) {
            std::cout << colName << " = " << value << std::endl;

            if (! doNoUpdates) {
                if (newRow.get() == 0) {
                    newRow.reset(dbt->NewRow());
                    if (newRow.get() == 0) {return false;}
                }
                newRow->Set(colName, value);
            }
        }
    }
    if (newRow.get() != 0) {
        newRow->Set(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                    subsysCfgId);
        dbt->AddRow(newRow.get());
    }

    return true;
}

bool _getDistinctDCMNames(boost::shared_ptr<NOVADB::Table> dbt,
                          const std::vector<int64_t>& subsystemConfigIdList,
                          std::vector<std::string>& dcmNameList)
{
    if (subsystemConfigIdList.size() == 0) {return false;}
    int64_t minId = subsystemConfigIdList[0];
    int64_t maxId = subsystemConfigIdList[0];
    for (int idx = 1; idx < (int) subsystemConfigIdList.size(); ++idx) {
        if (subsystemConfigIdList[idx] < minId) {
            minId = subsystemConfigIdList[idx];
        }
        if (subsystemConfigIdList[idx] > maxId) {
            maxId = subsystemConfigIdList[idx];
        }
    }

    NOVADB::Row* rowPtr;
    NOVADB::Column* columnPtr;
    dbt->Clear();
    dbt->AddDistinctColumn(DBColumnParams::DCMID_COLUMN_NAME);
    dbt->SetValidityRange(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                          minId, maxId);
    if (!dbt->LoadFromDB()) {return false;}
    for (int idx = 0; idx < dbt->NRow(); ++idx) {
        rowPtr = dbt->GetRow(idx);
        columnPtr = rowPtr->Col(DBColumnParams::DCMID_COLUMN_NAME);
        std::string dcmName;
        if (! columnPtr->Get(dcmName)) {
            return false;
        }
        //std::cout << "DCM Name = " << dcmName << std::endl;
        dcmNameList.push_back(dcmName);
    }
    return true;
}

bool _getDistinctTDUNames(boost::shared_ptr<NOVADB::Table> dbt,
                          const std::vector<int64_t>& subsystemConfigIdList,
                          std::vector<std::string>& tduNameList)
{
    if (subsystemConfigIdList.size() == 0) {return false;}
    int64_t minId = subsystemConfigIdList[0];
    int64_t maxId = subsystemConfigIdList[0];
    for (int idx = 1; idx < (int) subsystemConfigIdList.size(); ++idx) {
        if (subsystemConfigIdList[idx] < minId) {
            minId = subsystemConfigIdList[idx];
        }
        if (subsystemConfigIdList[idx] > maxId) {
            maxId = subsystemConfigIdList[idx];
        }
    }

    NOVADB::Row* rowPtr;
    NOVADB::Column* columnPtr;
    dbt->Clear();
    dbt->AddDistinctColumn(DBColumnParams::TDUNAME_COLUMN_NAME);
    dbt->SetValidityRange(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                          minId, maxId);
    if (!dbt->LoadFromDB()) {return false;}
    for (int idx = 0; idx < dbt->NRow(); ++idx) {
        rowPtr = dbt->GetRow(idx);
        columnPtr = rowPtr->Col(DBColumnParams::TDUNAME_COLUMN_NAME);
        std::string tduName;
        if (! columnPtr->Get(tduName)) {
            return false;
        }
        //std::cout << "TDU Name = " << tduName << std::endl;
        tduNameList.push_back(tduName);
    }
    return true;
}

bool _getDistinctHostNames(boost::shared_ptr<NOVADB::Table> dbt,
                           const std::vector<int64_t>& subsystemConfigIdList,
                           std::vector<std::string>& hostNameList)
{
    if (subsystemConfigIdList.size() == 0) {return false;}
    int64_t minId = subsystemConfigIdList[0];
    int64_t maxId = subsystemConfigIdList[0];
    for (int idx = 1; idx < (int) subsystemConfigIdList.size(); ++idx) {
        if (subsystemConfigIdList[idx] < minId) {
            minId = subsystemConfigIdList[idx];
        }
        if (subsystemConfigIdList[idx] > maxId) {
            maxId = subsystemConfigIdList[idx];
        }
    }

    NOVADB::Row* rowPtr;
    NOVADB::Column* columnPtr;
    dbt->Clear();
    dbt->AddDistinctColumn(DBColumnParams::HOSTNAME_COLUMN_NAME);
    dbt->SetValidityRange(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                          minId, maxId);
    if (!dbt->LoadFromDB()) {return false;}
    for (int idx = 0; idx < dbt->NRow(); ++idx) {
        rowPtr = dbt->GetRow(idx);
        columnPtr = rowPtr->Col(DBColumnParams::HOSTNAME_COLUMN_NAME);
        std::string hostName;
        if (! columnPtr->Get(hostName)) {
            return false;
        }
        //std::cout << "HOST Name = " << hostName << std::endl;
        hostNameList.push_back(hostName);
    }
    return true;
}

bool _updateTable(bool doNoUpdates,
                  const std::string& tableDefinitionFileName,
                  const std::vector<int64_t>& subsystemConfigIdList,
                  const int64_t& newSubsysCfgId)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    dbt.reset(new NOVADB::Table(tableDefinitionFileName.c_str()));
    std::string detName = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    std::cout << std::endl;
    std::cout << "*** Table = " << dbt->Name() << std::endl;

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }

    ConfigDataTree dataTree(tableDefinitionFileName,
                            subsystemConfigIdList);
    if (! dataTree.hasValidData()) {return false;}

    std::vector<std::string> selectorNameList =
        dataTree.getSelectorColumnNames();
    std::vector<std::string> dataColumnNameList =
        dataTree.getDataColumnNames();

    std::vector<std::string>::iterator nameIter;
    bool foundDCMSelector = false;
    nameIter = std::find(selectorNameList.begin(), selectorNameList.end(),
                         DBColumnParams::DCMID_COLUMN_NAME);
    if (nameIter != selectorNameList.end()) {
        foundDCMSelector = true;
    }
    bool foundFEBSelector = false;
    nameIter = std::find(selectorNameList.begin(), selectorNameList.end(),
                         DBColumnParams::FEBID_COLUMN_NAME);
    if (nameIter != selectorNameList.end()) {
        foundFEBSelector = true;
    }
    bool foundPixelSelector = false;
    nameIter = std::find(selectorNameList.begin(), selectorNameList.end(),
                         DBColumnParams::PIXELID_COLUMN_NAME);
    if (nameIter != selectorNameList.end()) {
        foundPixelSelector = true;
    }
    bool foundCoolingState = false;
    nameIter = std::find(selectorNameList.begin(), selectorNameList.end(),
                         "coolingstate");
    if (nameIter != selectorNameList.end()) {
        foundCoolingState = true;
    }
    bool foundPVName = false;
    nameIter = std::find(selectorNameList.begin(), selectorNameList.end(),
                         "pvname");
    if (nameIter != selectorNameList.end()) {
        foundPVName = true;
    }
    bool foundHostNameSelector = false;
    nameIter = std::find(selectorNameList.begin(), selectorNameList.end(),
                         DBColumnParams::HOSTNAME_COLUMN_NAME);
    if (nameIter != selectorNameList.end()) {
        foundHostNameSelector = true;
    }

    if (foundDCMSelector) {
        if (! dbt->GetConnection()) {return false;}
        if (! dbt->ExistsInDB()) {return false;}

        std::vector<std::string> dcmNameList;
        if (! _getDistinctDCMNames(dbt, subsystemConfigIdList, dcmNameList)) {
            return false;
        }
        dbt->Clear();

        if (foundFEBSelector && foundCoolingState) {
            _populateTableByDCMFEBAndThirdSelector(doNoUpdates, dbt,
                                                   newSubsysCfgId, dataTree,
                                                   "coolingstate",
                                                   dataColumnNameList,
                                                   dcmNameList);
        }
        else if (foundFEBSelector && foundPVName) {
            _populateTableByDCMFEBAndThirdSelector(doNoUpdates, dbt,
                                                   newSubsysCfgId, dataTree,
                                                   "pvname",
                                                   dataColumnNameList,
                                                   dcmNameList);
        }
        else if (foundFEBSelector && foundPixelSelector) {
            _populateTableByDCMFEBAndPixel(doNoUpdates,
                                           dbt, newSubsysCfgId, dataTree,
                                           dataColumnNameList, dcmNameList);
        }
        else if (foundFEBSelector) {
            _populateTableByDCMAndFEB(doNoUpdates,
                                      dbt, newSubsysCfgId, dataTree,
                                      dataColumnNameList, dcmNameList);
        }
        else {
            _populateTableByDCM(doNoUpdates,
                                dbt, newSubsysCfgId, dataTree,
                                dataColumnNameList, dcmNameList);
        }
    }
    else if (foundHostNameSelector) {
        if (! dbt->GetConnection()) {return false;}
        if (! dbt->ExistsInDB()) {return false;}

        std::vector<std::string> hostNameList;
        if (! _getDistinctHostNames(dbt, subsystemConfigIdList, hostNameList)) {
            return false;
        }
        dbt->Clear();

        _populateTableByHost(doNoUpdates,
                             dbt, newSubsysCfgId, dataTree,
                             dataColumnNameList, hostNameList);
    }
    else {
        _populateTable(doNoUpdates,
                       dbt, newSubsysCfgId, dataTree, dataColumnNameList);
    }

    if (! dbt->BulkInsertInDB()) {return false;}
    dbt->Clear();

    return true;
}

bool _updateTDUDelayTable(bool doNoUpdates,
                  const std::string& tableDefinitionFileName,
                  const std::vector<int64_t>& subsystemConfigIdList,
                  const int64_t& newSubsysCfgId)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    dbt.reset(new NOVADB::Table(tableDefinitionFileName.c_str()));
    std::string detName = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    std::cout << std::endl;
    std::cout << "*** Table = " << dbt->Name() << std::endl;

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }

    std::vector<std::string> selectorNameList;
    selectorNameList.push_back(DBColumnParams::TDUNAME_COLUMN_NAME);
    selectorNameList.push_back(DBColumnParams::TDUDELAYTYPE_COLUMN_NAME);
    std::vector<std::string> dataColumnNameList;
    dataColumnNameList.push_back("delay");

    ConfigDataTree dataTree(tableDefinitionFileName,
                            subsystemConfigIdList);
    if (! dataTree.hasValidData()) {return false;}
 
    if (! dbt->GetConnection()) {return false;}
    if (! dbt->ExistsInDB()) {return false;}

    std::vector<std::string> tduNameList;
    if (! _getDistinctTDUNames(dbt, subsystemConfigIdList, tduNameList)) {
        return false;
    }
    dbt->Clear();

    _populateTableByTDUAndDelayType(doNoUpdates, dbt, newSubsysCfgId,
                                    dataTree, dataColumnNameList, tduNameList);

    if (! dbt->BulkInsertInDB()) {return false;}
    dbt->Clear();

    return true;
}

bool _updateMultiRowNoInheritanceTable(
                    bool doNoUpdates,
                    const std::string& tableDefinitionFileName,
                    const std::vector<int64_t>& subsystemConfigIdList,
                    const int64_t& newSubsysCfgId)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    dbt.reset(new NOVADB::Table(tableDefinitionFileName.c_str()));
    std::string detName = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    std::vector<std::string> colNames = dbt->GetColNames();

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }

    if (! dbt->GetConnection()) {return false;}
    std::cout << std::endl;
    std::cout << "*** Table = " << dbt->Name() << std::endl;

    for (uint32_t cfgIdx = 0; cfgIdx<subsystemConfigIdList.size(); ++cfgIdx) {
        dbt->Clear();
        dbt->SetValidityRange(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                              subsystemConfigIdList[cfgIdx]);
        if (! dbt->LoadFromDB()) {
            return false;
        }

        if (dbt->NRow() > 0) {
            int initialRowCount = dbt->NRow();
            for (int iRow = 0; iRow < initialRowCount; ++iRow) {
                boost::shared_ptr<Row> newRow(dbt->NewRow());
                Row* rowPtr = dbt->GetRow(iRow);

                bool firstPrint = true;
                for (uint32_t idx = 0; idx < colNames.size(); ++idx) {

                    std::string colName = boost::to_lower_copy(colNames[idx]);
                    if (DBColumnUtils::isInfrastructureColumn(colName)) {continue;}
                    if (DBColumnUtils::isSubsystemConfigReferenceColumn(colName)) {
                        continue;
                    }

                    std::string value;
                    Column* colPtr = rowPtr->Col(colName);
                    if (colPtr->IsNull()) {continue;}
                    if (! colPtr->Get(value)) {return false;}

                    std::string colType = colPtr->Type();
                    if (colType == "bool" || colType == "boolean") {
                        if (value == "t" || value == "1") {
                            newRow->Set(colName, true);
                        }
                        else {
                            newRow->Set(colName, false);
                        }
                    }
                    else {
                        newRow->Set(colName, value);
                    }

                    if (firstPrint) {firstPrint = false;}
                    else {std::cout << ", ";}
                    std::cout << colName << " = " << value;
                }
                std::cout << std::endl;

                if (! doNoUpdates) {
                    newRow->Set(NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME,
                                newSubsysCfgId);
                    dbt->AddRow(newRow.get());
                    if (! dbt->WriteToDB()) {return false;}
                }
            }

            break;
        }
    }

    return true;
}

/**
 * Command-line utility for flattening an existing DAQ Configuration dataset.
 */

int main(int argc, char **argv)
{
    GlobalCleanupHandler globalCleanupHandler;

    // parse command line arguments and print usage hints, if needed.
    std::string timestamp = "";
    bool printUsage = false;
    bool noUpdate = false;
    bool forceUpdate = false;
    int c;
    while ((c = getopt (argc, argv, "nhf")) != -1) {
        switch (c)
        {
        case 'h':
            printUsage = true;
            break;
        case 'n':
            noUpdate = true;
            break;
        case 'f':
            forceUpdate = true;
            break;
        default:
            printUsage = true;
            break;
        }
    }
    globalCleanupHandler.noUpdate = noUpdate;

    std::string daqOrSlowControls;
    int64_t inputGlobalConfigId = 0;
    std::string configName;
    std::string dcmOpMode;
    std::string detectorName;
    if (! printUsage) {
        if ((argc - optind) < 2) {
            printUsage = true;
        }
        else {
            {
                std::string tmpString(argv[optind]);
                if (tmpString == "DAQ" || tmpString == "SlowControls") {
                    daqOrSlowControls = tmpString;
                }
                else {
                    printUsage = true;
                }
                ++optind;
            }
            try {
                std::string firstArg(argv[optind]);
                inputGlobalConfigId = boost::lexical_cast<int64_t>(firstArg);
                ++optind;
            }
            catch (...) {
                if ((argc - optind) < 3) {
                    printUsage = true;
                }
                else {
                    {
                        std::string tmpString(argv[optind]);
                        configName = tmpString;
                        ++optind;
                    }
                    {
                        std::string tmpString(argv[optind]);
                        dcmOpMode = tmpString;
                        ++optind;
                    }
                    {
                        std::string tmpString(argv[optind]);
                        detectorName = tmpString;
                        ++optind;
                    }
                }
            }
        }
    }

    // print usage hints, if needed.
    if (printUsage) {
        std::cout << std::endl;
        std::cout << "Usage: " << argv[0] << " [-n] [-f]"
                  << " <\"DAQ\"|\"SlowControls\"> <globalCfgName> "
                  << "<dcmOpMode> <detectorName>" << std::endl;
        std::cout << " OR " << std::endl;
        std::cout << "       " << argv[0]
                  << " <\"DAQ\"|\"SlowControls\"> <globalCfgId>" << std::endl;
        std::cout << "where" << std::endl;
        std::cout << " * \"DAQ\"|\"SlowControls\" selects which set "
                  << "of database tables to use" << std::endl;
        std::cout << " * globalCfgName specifies the "
                  << "named configuration to flatten" << std::endl;
        std::cout << " * dcmOpMode specifies the DCM operating mode "
                  << "(e.g. DCS)" << std::endl;
        std::cout << " * detectorName specifies the detector name "
                  << "(e.g. NDOS)" << std::endl;
        std::cout << " * globalCfgId specifies the "
                  << "named global configuration to show" << std::endl;
        std::cout << " * -n specifies that no updates should be done "
                  << "(console messages are still printed)" << std::endl;
        std::cout << " * -f specifies that DB updates should be done "
                  << "(forced) even if" << std::endl;
        std::cout << "   there are changes that might "
                  << "be overwritten" << std::endl;
        std::cout << "   --> use the -f option only when "
                  << "ABSOLUTELY NECESSARY" << std::endl;
        std::cout << "Examples:" << std::endl;
        std::cout << "   " << argv[0] << " InitialGlobalConfig "
                  << "DCS NDOS" << std::endl;
        std::cout << "   " << argv[0] << " 62" << std::endl;
        return -1;
    }
    boost::shared_ptr<NOVADB::Table> dbt;

    std::string dbName =
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBNAME");
    std::cout << "Are you sure that you want to flatten the inheritance hierarchy "
         << "in the " << dbName << " database (yes/no): ";
    std::string response;
    std::cin >> response;
    if (response.length() == 0 ||
        (response[0] != 'y' && response[0] != 'Y')) {
        return 1;
    }

    // fetch the global config ID, if needed
    int64_t globalConfigId = inputGlobalConfigId;
    if (globalConfigId == 0) {
        GlobalNameSpec nameSpec(configName, dcmOpMode, detectorName);
        nameSpec.isDAQConfiguration(daqOrSlowControls == "DAQ");
        if (! NamedConfigUtils::getGlobalConfigId(nameSpec, globalConfigId)) {
            std::cout << std::endl << "No global configuration was found "
                      << "for the parameters that you specified!"
                      << std::endl;
            std::cout << "Please double-check that the values for the input "
                      << "parameters are correct..."
                      << std::endl;
            return 2;
        }
    }
    std::cout << "The source global configuration ID = " << globalConfigId
              << std::endl;

    // copy the global configuration to a new one
    IDSpec globalIdSpec(globalConfigId, IDSpec::GLOBAL,
                        (daqOrSlowControls == "DAQ" ?
                         IDSpec::DAQ : IDSpec::SLOWCONTROLS));
    int64_t newGlobalConfigId = -1;
    if (! noUpdate) {
        dbutils::SimpleParameterSet paramSet;
        paramSet.put(NamedConfigUtils::DESCRIPTION_COLUMN_NAME,
                     "Result of collapsing (flattening) the ancestry tree.");
        novadaq::Status resultStatus =
            NamedConfigUtils::copyGlobalConfig(globalIdSpec, false, paramSet,
                                               "", forceUpdate,
                                               newGlobalConfigId);
        if (resultStatus.code ==
            NamedConfigUtils::COPY_FOUND_UPDATE_IN_PROGRESS) {
            std::cout << "Unable to create a copy of the named "
                      << "global configuration with ID = " << globalConfigId
                      << ": the creation of a new configuration is "
                      << "already in progress ("
                      << resultStatus.message << ")"
                      << std::endl;
            return 2;
        }
        else if (resultStatus.code ==
                 NamedConfigUtils::COPY_FOUND_INTERMEDIATE_UPDATE) {
            std::cout << "Unable to create a copy of the named "
                      << "global configuration with ID = " << globalConfigId
                      << ": intermediate changes would be overwritten by "
                      << "this change (" << resultStatus.message << ")"
                      << std::endl;
            return 2;
        }
        else if (resultStatus.code != NamedConfigUtils::COPY_SUCCEEDED) {
            std::cout << "Unable to create a copy of the named "
                      << "global configuration with ID = " << globalConfigId
                      << ": " << resultStatus.message
                      << std::endl;
            return 2;
        }
        std::cout << "The new global configuration ID = "
                  << newGlobalConfigId << std::endl;
        globalCleanupHandler.daqOrSlowControls = daqOrSlowControls;
        globalCleanupHandler.globalConfigId = newGlobalConfigId;
    }
    

    // fetch the list of subsystem config IDs
    std::vector<int64_t> subsystemConfigIdList;
    std::vector<SubsystemNameSpec> subsystemNameSpecList;
    if (! NamedConfigUtils::
        getSubsystemConfigIdsFromGlobalId(globalIdSpec,
                                          subsystemConfigIdList,
                                          subsystemNameSpecList)) {
        std::cout << "Unable to find the subsystem configurations for "
                  << "global config ID = " << globalConfigId
                  << std::endl;
        return 2;
    }

    // deal with each subsystem
    bool foundDCS = false;
    std::string fname;
    for (uint32_t idx = 0; idx < subsystemConfigIdList.size(); ++idx) {
        SubsystemCleanupHandler subsystemCleanupHandler;
        subsystemCleanupHandler.noUpdate = noUpdate;

        std::cout << std::endl;
        std::cout << "Subsystem:";
        std::cout << " ID=" << subsystemConfigIdList[idx];
        std::cout << ", name="
                  << subsystemNameSpecList[idx].getSubsystemName();
        std::cout << ", configType="
                  << subsystemNameSpecList[idx].getConfigType();
        std::cout << std::endl;

        IDSpec subsystemIdSpec(subsystemConfigIdList[idx],
                               IDSpec::SUBSYSTEM,
                               (daqOrSlowControls == "DAQ" ?
                                IDSpec::DAQ : IDSpec::SLOWCONTROLS));
        std::vector<int64_t> subsystemConfigIdLineage;
        if (! NamedConfigUtils::
            getConfigIdAncestry(subsystemIdSpec, subsystemConfigIdLineage)) {
            std::cout << "Unable to find the subsystem configuration "
                      << "ancestry for subsystem config ID = "
                      << subsystemConfigIdList[idx] << std::endl;
            return 2;
        }

        std::cout << "  Ancestor IDs:";
        for (uint32_t jdx = 0; jdx < subsystemConfigIdLineage.size(); ++jdx) {
            std::cout << " " << subsystemConfigIdLineage[jdx];
        }
        std::cout << std::endl;

        if (subsystemNameSpecList[idx].getSubsystemName() ==
            SubsystemNameSpec::DCM_SUBSYSTEM_NAME &&
            subsystemNameSpecList[idx].getConfigType() ==
            SubsystemNameSpec::HARDWARE_CONFIG_TYPE) {

            if (foundDCS) {
                std::cout << "Skipping DCMApplication hardware CFG "
                          << "associated with DCS CFG data." << std::endl;
                continue;
            }
        }

        // copy the subsystem configuration to a new one
        int64_t newSubsystemConfigId = -1;
        if (! noUpdate) {
            dbutils::SimpleParameterSet paramSet;
            paramSet.put(NamedConfigUtils::DESCRIPTION_COLUMN_NAME,
                         "Result of collapsing (flattening) the ancestry tree.");
            paramSet.put(NamedConfigUtils::GLOBALCFGID_COLUMN_NAME,
                         boost::lexical_cast<std::string>(newGlobalConfigId));
            novadaq::Status resultStatus = NamedConfigUtils::
                copySubsystemConfig(subsystemIdSpec, false, paramSet,
                                    "", forceUpdate, newSubsystemConfigId);
            if (resultStatus.code ==
                NamedConfigUtils::COPY_FOUND_UPDATE_IN_PROGRESS) {
                std::cout << "Unable to create a copy of the named "
                          << "subsystem configuration with ID = "
                          << subsystemConfigIdList[idx]
                          << ": the creation of a new configuration is "
                          << "already in progress ("
                          << resultStatus.message << ")"
                          << std::endl;
                return 2;
            }
            else if (resultStatus.code ==
                     NamedConfigUtils::COPY_FOUND_INTERMEDIATE_UPDATE) {
                std::cout << "Unable to create a copy of the named "
                          << "subsystem configuration with ID = "
                          << subsystemConfigIdList[idx]
                          << ": intermediate changes would be overwritten by "
                          << "this change (" << resultStatus.message << ")"
                          << std::endl;
                return 2;
            }
            else if (resultStatus.code != NamedConfigUtils::COPY_SUCCEEDED) {
                std::cout << "Unable to create a copy of the named "
                          << "subsystem configuration with ID = "
                          << subsystemConfigIdList[idx]
                          << ": " << resultStatus.message
                          << std::endl;
                return 2;
            }


            if (resultStatus.code != NamedConfigUtils::COPY_SUCCEEDED) {
                std::cout << "Unable to create a copy of the named "
                          << "subsystem configuration with ID = "
                          << subsystemConfigIdList[idx] << std::endl;
                return 2;
            }
            std::cout << "  The new subsystem configuration ID = "
                      << newSubsystemConfigId << std::endl;
            subsystemCleanupHandler.daqOrSlowControls = daqOrSlowControls;
            subsystemCleanupHandler.subsystemConfigId = newSubsystemConfigId;
        }

        // DCM Hardware configuration
        if (subsystemNameSpecList[idx].getSubsystemName() ==
            SubsystemNameSpec::DCM_SUBSYSTEM_NAME &&
            subsystemNameSpecList[idx].getConfigType() ==
            SubsystemNameSpec::HARDWARE_CONFIG_TYPE) {

            // system-wide parameters
            fname = "NovaDatabase/tables/DAQConfig/DCMSystemParameters.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // dcm-based parameters
            fname = "NovaDatabase/tables/DAQConfig/DCMApplicationParameters.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/DCMDataDevParameters.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/DCMFPGAFirmwareLocations.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/DCMFPGAParameters.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/FEBEnableMasks.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/FEBFPGAFirmwareLocations.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/TimingSystemSettings.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/DCMTimingDelaySettings.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // feb-based parameters
            fname = "NovaDatabase/tables/DAQConfig/ASICRegisterSettings.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/DSODataRegulatorSettings.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/FEBPulserParameters.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/PixelEnableMasks.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // pixel-base parameters
            fname = "NovaDatabase/tables/DAQConfig/PixelOffsets.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/PixelThresholds.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
        }

        // DCM Run configuration
        else if (subsystemNameSpecList[idx].getSubsystemName() ==
                 SubsystemNameSpec::DCM_SUBSYSTEM_NAME &&
                 subsystemNameSpecList[idx].getConfigType() ==
                 SubsystemNameSpec::RUN_CONFIG_TYPE) {

            // system-wide parameters
            fname = "NovaDatabase/tables/DAQConfig/DCMSystemRunParameters.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // dcm-based parameters
            fname = "NovaDatabase/tables/DAQConfig/DCMApplicationRunParameters.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            fname = "NovaDatabase/tables/DAQConfig/DCMDataDevRunParameters.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
        }

        // DCM Connections configuration
        else if (subsystemNameSpecList[idx].getSubsystemName() ==
                 SubsystemNameSpec::DCM_SUBSYSTEM_NAME &&
                 subsystemNameSpecList[idx].getConfigType() ==
                 SubsystemNameSpec::CONNECT_CONFIG_TYPE) {

            // dcm-based parameters
            fname = "NovaDatabase/tables/DAQConfig/DCMApplicationConnectParameters.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
        }

        // DL Run configuration
        else if (subsystemNameSpecList[idx].getSubsystemName() ==
                 SubsystemNameSpec::DL_SUBSYSTEM_NAME &&
                 subsystemNameSpecList[idx].getConfigType() ==
                 SubsystemNameSpec::RUN_CONFIG_TYPE) {

            // streams
            fname = "NovaDatabase/tables/DAQConfig/DataLoggerStreams.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // application parameters
            fname = "NovaDatabase/tables/DAQConfig/DataLoggerSystemParameters.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
        }

        // DaqMonitor Run configuration
        else if (subsystemNameSpecList[idx].getSubsystemName() ==
                 SubsystemNameSpec::DAQMONITOR_SUBSYSTEM_NAME &&
                 subsystemNameSpecList[idx].getConfigType() ==
                 SubsystemNameSpec::RUN_CONFIG_TYPE) {

            // run config
            fname = "NovaDatabase/tables/DAQConfig/DaqMonitorRunParameters.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

        }

        // GT Run configuration
        else if (subsystemNameSpecList[idx].getSubsystemName() ==
                 SubsystemNameSpec::GT_SUBSYSTEM_NAME &&
                 subsystemNameSpecList[idx].getConfigType() ==
                 SubsystemNameSpec::RUN_CONFIG_TYPE) {

            // GT general parameters
            fname = "NovaDatabase/tables/DAQConfig/GTGeneral.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }


            // calibration triggers
            fname = "NovaDatabase/tables/DAQConfig/CalibrationTriggers.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // external triggers
            fname = "NovaDatabase/tables/DAQConfig/ExternalTriggers.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // datadriven_triggers
            fname = "NovaDatabase/tables/DAQConfig/DataDrivenTriggers.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // daqstatus_triggers
            fname = "NovaDatabase/tables/DAQConfig/DAQStatusTriggers.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // manual triggers
            fname = "NovaDatabase/tables/DAQConfig/ManualTriggers.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // spill triggers
            fname = "NovaDatabase/tables/DAQConfig/SpillTriggers.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // SNEWS triggers
            fname = "NovaDatabase/tables/DAQConfig/SNEWSTriggers.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
           
	    // SuperNova triggers
            fname = "NovaDatabase/tables/DAQConfig/SuperNovaTrigger.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
            
	    // SuperNova filters
            fname = "NovaDatabase/tables/DAQConfig/SuperNovaFilter.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
            // GlobalTrigger Queue
            fname = "NovaDatabase/tables/DAQConfig/GTQueue.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
            //SNEWS Pipe
            fname = "NovaDatabase/tables/DAQConfig/SNEWSPipe.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

	    // Global Throttle
            fname = "NovaDatabase/tables/DAQConfig/GlobalThrottle.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // DDT Throttle
            fname = "NovaDatabase/tables/DAQConfig/DDTThrottle.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // trigger offsets
            fname = "NovaDatabase/tables/DAQConfig/TriggerOffsets.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
        }

        // DDT configuration
        else if (subsystemNameSpecList[idx].getSubsystemName() ==
                 SubsystemNameSpec::DDT_SUBSYSTEM_NAME &&
                 subsystemNameSpecList[idx].getConfigType() ==
                 SubsystemNameSpec::CONNECT_CONFIG_TYPE) {

            // DDTManager connection parameters
            fname = "NovaDatabase/tables/DAQConfig/DDTManagerConnectParameters.xml";
	    // this is an exception table, so we need to use inheritance
            // if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
            //                                         subsystemConfigIdLineage,
            //                                         newSubsystemConfigId))
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) 
	    {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
	}

        // BNEVB Run configuration
        else if (subsystemNameSpecList[idx].getSubsystemName() ==
                 SubsystemNameSpec::BNEVB_SUBSYSTEM_NAME &&
                 subsystemNameSpecList[idx].getConfigType() ==
                 SubsystemNameSpec::RUN_CONFIG_TYPE) {

            // application parameters
            fname = "NovaDatabase/tables/DAQConfig/BNEVBRunParameters.xml";
            if (! _updateMultiRowNoInheritanceTable(noUpdate, fname,
                                                    subsystemConfigIdLineage,
                                                    newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
        }

        // slow controls configuration
        else if (subsystemNameSpecList[idx].getSubsystemName() ==
                 "APDHardware" &&
                 subsystemNameSpecList[idx].getConfigType() ==
                 "slowcontrols") {
            foundDCS = true;

            // temperatures
            fname = "NovaDatabase/tables/DCS/APDTemperatureSettings.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // temperature enables
            fname = "NovaDatabase/tables/DCS/DCSFEBEnableMasks.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // voltages
            fname = "NovaDatabase/tables/DCS/APDHighVoltages.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // FEB dry air channels
            fname = "NovaDatabase/tables/DCS/FEBDryAirChannels.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }

            // alarm thresholds
            fname = "NovaDatabase/tables/DCS/EPICSAlarmThresholds.xml";
            if (! _updateTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
        }

        // TDUManager hardware configuration
        else if (subsystemNameSpecList[idx].getSubsystemName() ==
                 SubsystemNameSpec::TDUMANAGER_SUBSYSTEM_NAME &&
                 subsystemNameSpecList[idx].getConfigType() ==
                 SubsystemNameSpec::HARDWARE_CONFIG_TYPE) {

            // timing delay constants
            fname = "NovaDatabase/tables/DAQConfig/TDUTimingDelaySettings.xml";
            if (! _updateTDUDelayTable(noUpdate, fname, subsystemConfigIdLineage,
                               newSubsystemConfigId)) {
                std::cout << "Unable to update " << fname << std::endl;
                return 2;
            }
        }

        else {
            std::cout << "ERROR: unknown combination of subsystem "
                      << "and configuration type: "
                      << subsystemNameSpecList[idx].getSubsystemName()
                      << " & " << subsystemNameSpecList[idx].getConfigType()
                      << std::endl;
            return 3;
        }

        subsystemCleanupHandler.success = true;
    }

    globalCleanupHandler.success = true;

    return 0;
}
