#include <DatabaseUtils/DAQConfig/NDMRunConfig.h>
#include <DatabaseUtils/DAQConfig/ConfigDataTree.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/DBColumnUtils.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <NovaDatabase/Table.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <boost/algorithm/string.hpp>

namespace NOVADB = nova::database;

namespace dbutils {
namespace daqconfig {

/**
 * Creates a DaqMonitor (aka NDM) run configuration object from the specified
 * list of subsystem configuration IDs.
 *
 * @param subsystemConfigIdList The list of subsystem
 *        configuration IDs.
 */
NDMRunConfig::
NDMRunConfig(const std::vector<int64_t>& subsystemConfigIdList) : 
  _rrdRootDir(""), _thresholdDir("")
{
    _primarySubsystemConfigId = -1;
    if (subsystemConfigIdList.size() > 0) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a NDMRunConfig run configuration object from the specified
 * subsystem configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 */
NDMRunConfig::NDMRunConfig(const SubsystemNameSpec& nameSpec) :
  _rrdRootDir(""),_thresholdDir("")
{
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestry(nameSpec, subsystemConfigIdList)) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a NDMRunConfig run configuration object from the specified
 * global configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 */
NDMRunConfig::NDMRunConfig(const GlobalNameSpec& nameSpec) :
  _rrdRootDir(""),_thresholdDir("")
{
    _primarySubsystemConfigId = -1;
    int64_t subsystemConfigId;
    bool status = NamedConfigUtils::
        getSubsystemConfigIdFromGlobalName(nameSpec,
                                           SubsystemNameSpec::
                                           DAQMONITOR_SUBSYSTEM_NAME,
                                           SubsystemNameSpec::
                                           RUN_CONFIG_TYPE,
                                           subsystemConfigId);

    std::vector<int64_t> subsystemConfigIdList;
    if (status) {
        IDSpec idSpec(subsystemConfigId, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        status = NamedConfigUtils::
            getConfigIdAncestry(idSpec, subsystemConfigIdList);
    }

    if (status) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a NDMRunConfig run configuration object from the specified
 * global configuration ID.
 *
 * @param globalConfigId The ID of the global configuration of interest.
 */
NDMRunConfig::NDMRunConfig(const int64_t& globalConfigId) :
  _rrdRootDir(""),_thresholdDir("")
{
    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL, IDSpec::DAQ);
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestryFromGlobalId(idSpec,
                                                 SubsystemNameSpec::
                                                 DAQMONITOR_SUBSYSTEM_NAME,
                                                 SubsystemNameSpec::
                                                 RUN_CONFIG_TYPE,
                                                 subsystemConfigIdList)) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList);
    }
    else {
        _hasValidData = false;
    }
}

bool NDMRunConfig::
_fetchAllData(const std::vector<int64_t>& subsystemConfigIdList)
{
    ConfigDataSelector cfgSelector;
    
    std::vector<std::string> dataColumnsOfInterest;
    dataColumnsOfInterest.push_back("rrdRootDir");
    dataColumnsOfInterest.push_back("thresholdDir");

    ConfigDataTree
      daqMonitorParamDataSet("NovaDatabase/tables/DAQConfig/DaqMonitorRunParameters.xml",
			     subsystemConfigIdList, cfgSelector,
			     dataColumnsOfInterest);

    if ( ! daqMonitorParamDataSet.getTrueValue("rrdRootDir", _rrdRootDir) ) return false;
    if ( ! daqMonitorParamDataSet.getTrueValue("thresholdDir",_thresholdDir) ) return false;

    return true;
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
