#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <DatabaseUtils/DAQConfig/DBColumnUtils.h>
#include <NovaDatabase/Util.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <NovaDAQConventions/DAQConventions.h>

#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <iostream>

namespace NOVADB = nova::database;

namespace dbutils {
namespace daqconfig {

const std::string NamedConfigUtils::
DEFAULT_GLOBAL_NAMES_TABLE_FILE("NovaDatabase/tables/DAQConfig/NamedGlobalConfigurations.xml");
const std::string NamedConfigUtils::
DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE("NovaDatabase/tables/DAQConfig/NamedSubsystemConfigurations.xml");
const std::string NamedConfigUtils::
DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE("NovaDatabase/tables/DCS/DCSNamedGlobalConfigurations.xml");
const std::string NamedConfigUtils::
DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE("NovaDatabase/tables/DCS/DCSNamedSubsystemConfigurations.xml");

const std::string NamedConfigUtils::ID_COLUMN_NAME("id");
const std::string NamedConfigUtils::NAME_COLUMN_NAME("name");
const std::string NamedConfigUtils::DCMOPMODE_COLUMN_NAME("dcmopmode");
const std::string NamedConfigUtils::DETECTORID_COLUMN_NAME("detectorname");
const std::string NamedConfigUtils::DETECTORNAME_COLUMN_NAME("detectorname");
const std::string NamedConfigUtils::DESCRIPTION_COLUMN_NAME("description");
const std::string NamedConfigUtils::SUBSYSTEMNAME_COLUMN_NAME("subsystemname");
const std::string NamedConfigUtils::CONFIGTYPE_COLUMN_NAME("configtype");
const std::string NamedConfigUtils::PARENTID_COLUMN_NAME("parentid");
const std::string NamedConfigUtils::GLOBALCFGID_COLUMN_NAME("globalcfgid");
const std::string NamedConfigUtils::SUBSYSCFGID_COLUMN_NAME("subsyscfgid");
const std::string NamedConfigUtils::USAGECOUNT_COLUMN_NAME("usagecount");
const std::string NamedConfigUtils::ISOBSOLETE_COLUMN_NAME("isobsolete");
const std::string NamedConfigUtils::ISCHANGING_COLUMN_NAME("ischanging");
const std::string NamedConfigUtils::ISCOMPLETE_COLUMN_NAME("iscomplete");
const std::string NamedConfigUtils::INSERTTIME_COLUMN_NAME("inserttime");
const std::string NamedConfigUtils::INSERTUSER_COLUMN_NAME("insertuser");
const std::string NamedConfigUtils::UPDATETIME_COLUMN_NAME("updatetime");
const std::string NamedConfigUtils::UPDATEUSER_COLUMN_NAME("updateuser");

const std::string NamedConfigUtils::NDOS_DETECTOR_NAME("NDOS");
const std::string NamedConfigUtils::NEAR_DETECTOR_NAME("NearDet");
const std::string NamedConfigUtils::FAR_DETECTOR_NAME("FarDet");

const std::string NamedConfigUtils::LATEST_TIME_STRING("Latest");

const int32_t NamedConfigUtils::COPY_SUCCEEDED = 11;
const int32_t NamedConfigUtils::COPY_FAILED = 12;
const int32_t NamedConfigUtils::COPY_FOUND_UPDATE_IN_PROGRESS = 13;
const int32_t NamedConfigUtils::COPY_FOUND_INTERMEDIATE_UPDATE = 14;

/**
 * Fetches the database ID for the specified named global configuration.
 *
 * NOTE that this method assumes certain column names for the database
 * table that contains the configuration names.  If any of these
 * assumptions are not valid for the given globalNamesTableFile,
 * then this method will always return false.
 *
 * @param globalNameSpec The specification of the named global
 *        configuration of interest.
 * @param globalConfigId The database ID of the named configuration
 *        [output parameter].
 *        (If the status returned by this method is false, then the value
 *        returned in this parameter is undefined.)
 * @param globalNamesTableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 *
 * @returns true if the lookup of the ID of the specified named
 *        configuration worked, false if not.
 */
bool NamedConfigUtils::
getGlobalConfigId(const GlobalNameSpec& globalNameSpec,
                  int64_t& globalConfigId,
                  const std::string& globalNamesTableFile)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    if (globalNameSpec.isSlowControlsConfiguration() &&
        globalNamesTableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(globalNamesTableFile.c_str()));
    }

    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    NamedConfigSelectionCriteria selections;
    selections[NAME_COLUMN_NAME] = globalNameSpec.getConfigName();
    selections[DCMOPMODE_COLUMN_NAME] = globalNameSpec.getDCMOpMode();
    selections[DETECTORID_COLUMN_NAME] = globalNameSpec.getDetectorId();
    if (globalNameSpec.hasTimestamp()) {
        selections[INSERTTIME_COLUMN_NAME] = globalNameSpec.getTimestamp();
    }
    if (! globalNameSpec.skipInvalidStateTest()) {
        selections[ISOBSOLETE_COLUMN_NAME] = "false";
        selections[ISCHANGING_COLUMN_NAME] = "false";
        selections[ISCOMPLETE_COLUMN_NAME] = "true";
    }

    return _getId(dbt, selections, globalConfigId);
}

/**
 * Fetches the database ID for the specified named subsystem configuration.
 *
 * NOTE that this method assumes certain column names for the database
 * table that contains the configuration names.  If any of these
 * assumptions are not valid for the given subsystemNamesTableFile,
 * then this method will always return false.
 *
 * @param subsystemNameSpec The specification of the named subsystem
 *        configuration of interest.
 * @param subsystemConfigId The database ID of the named configuration
 *        [output parameter].
 *        (If the status returned by this method is false, then the value
 *        returned in this parameter is undefined.)
 * @param subsystemNamesTableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 *
 * @returns true if the lookup of the ID of the specified named
 *        configuration worked, false if not.
 */
bool NamedConfigUtils::
getSubsystemConfigId(const SubsystemNameSpec& subsystemNameSpec,
                     int64_t& subsystemConfigId,
                     const std::string& subsystemNamesTableFile)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    if (subsystemNameSpec.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }

    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    NamedConfigSelectionCriteria selections;
    selections[NAME_COLUMN_NAME] = subsystemNameSpec.getConfigName();
    selections[DCMOPMODE_COLUMN_NAME] = subsystemNameSpec.getDCMOpMode();
    selections[DETECTORID_COLUMN_NAME] = subsystemNameSpec.getDetectorId();
    selections[SUBSYSTEMNAME_COLUMN_NAME]=subsystemNameSpec.getSubsystemName();
    selections[CONFIGTYPE_COLUMN_NAME] = subsystemNameSpec.getConfigType();
    if (subsystemNameSpec.hasTimestamp()) {
        selections[INSERTTIME_COLUMN_NAME] = subsystemNameSpec.getTimestamp();
    }
    selections[ISOBSOLETE_COLUMN_NAME] = "false";
    selections[ISCHANGING_COLUMN_NAME] = "false";
    selections[ISCOMPLETE_COLUMN_NAME] = "true";

    return _getId(dbt, selections, subsystemConfigId);
}

/**
 * Fetches the full set of database IDs for the specified named
 * configuration.  This includes the input ID
 * *and* all of the ancestor named configurations.
 *
 * The ID of the requested configuration is first in the returned
 * ID list, followed by its parent then grandparent, etc.  If the
 * requested configuration has no ancestors, then only the ID of
 * the requested configuration will be added to the ID list.
 *
 * NOTE that this method assumes certain column names for the database
 * table that contains the configuration names.  If any of these
 * assumptions are not valid for the table file that is given (or
 * inferred), then this method will always return false.
 *
 * @param idSpec The ID-based specification of the named configuration
 *        of interest.
 * @param idList The list of database IDs for the named configuration
 *        and its ancestors [output parameter].
 *        (If the status returned by this method is false, then the
 *        values returned in this parameter are undefined.)
 * @param tableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 *
 * @returns true if the lookup of the ID of the specified named
 *        configuration and its ancestors worked, false if not.
 */
bool NamedConfigUtils::
getConfigIdAncestry(const IDSpec& idSpec,
                    std::vector<int64_t>& idList,
                    const std::string& tableFile)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    if (tableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        if (idSpec.isGlobalConfiguration()) {
            if (idSpec.isDAQConfiguration()) {
                dbt.reset(new NOVADB::
                          Table(DEFAULT_GLOBAL_NAMES_TABLE_FILE.c_str()));
            }
            else {  // slow controls
                dbt.reset(new NOVADB::
                          Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
            }
        }
        else {  // subsystem
            if (idSpec.isDAQConfiguration()) {
                dbt.reset(new NOVADB::
                          Table(DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
            }
            else {  // slow controls
                dbt.reset(new NOVADB::
                          Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
            }
        }
    }
    else {
        dbt.reset(new NOVADB::Table(tableFile.c_str()));
    }

    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    idList.push_back(idSpec.getConfigId());
    uint32_t previousSize = idList.size() - 1;

    while (idList.size() > previousSize) {
        previousSize = idList.size();
        if (! _findNextAncestor(dbt, idList)) {
            return false;
        }
    }
    return true;
}

/**
 * Fetches the full set of database IDs for the specified named
 * configuration.  This includes the ID corresponding to the name
 * *and* all of the ancestor named configurations.
 *
 * The ID of the requested configuration is first in the returned
 * ID list, followed by its parent then grandparent, etc.  If the
 * requested configuration has no ancestors, then only the ID of
 * the requested configuration will be added to the ID list.
 *
 * NOTE that this method assumes certain column names for the database
 * table that contains the configuration names.  If any of these
 * assumptions are not valid for the given globalNamesTableFile,
 * then this method will always return false.
 *
 * @param globalNameSpec The specification of the named global
 *        configuration of interest.
 * @param idList The list of database IDs for the named configuration
 *        and its ancestors [output parameter].
 *        (If the status returned by this method is false, then the
 *        values returned in this parameter are undefined.)
 * @param globalNamesTableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 *
 * @returns true if the lookup of the ID of the specified named
 *        configuration and its ancestors worked, false if not.
 */
bool NamedConfigUtils::
getGlobalConfigIdAncestry(const GlobalNameSpec& globalNameSpec,
                          std::vector<int64_t>& idList,
                          const std::string& globalNamesTableFile)
{
    int64_t globalConfigId;

    boost::shared_ptr<NOVADB::Table> dbt;
    if (globalNameSpec.isSlowControlsConfiguration() &&
        globalNamesTableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(globalNamesTableFile.c_str()));
    }

    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    NamedConfigSelectionCriteria selections;
    selections[NAME_COLUMN_NAME] = globalNameSpec.getConfigName();
    selections[DCMOPMODE_COLUMN_NAME] = globalNameSpec.getDCMOpMode();
    selections[DETECTORID_COLUMN_NAME] = globalNameSpec.getDetectorId();
    if (globalNameSpec.hasTimestamp()) {
        selections[INSERTTIME_COLUMN_NAME] = globalNameSpec.getTimestamp();
    }
    if (! globalNameSpec.skipInvalidStateTest()) {
        selections[ISOBSOLETE_COLUMN_NAME] = "false";
        selections[ISCHANGING_COLUMN_NAME] = "false";
        selections[ISCOMPLETE_COLUMN_NAME] = "true";
    }

    if (! _getId(dbt, selections, globalConfigId)) {
        return false;
    }

    idList.push_back(globalConfigId);
    uint32_t previousSize = idList.size() - 1;

    while (idList.size() > previousSize) {
        previousSize = idList.size();
        if (! _findNextAncestor(dbt, idList)) {
            return false;
        }
    }
    return true;
}

/**
 * Fetches the full set of database IDs for the specified named
 * configuration.  This includes the ID corresponding to the name
 * *and* all of the ancestor named configurations.
 *
 * The ID of the requested configuration is first in the returned
 * ID list, followed by its parent then grandparent, etc.  If the
 * requested configuration has no ancestors, then only the ID of
 * the requested configuration will be added to the ID list.
 *
 * NOTE that this method assumes certain column names for the database
 * table that contains the configuration names.  If any of these
 * assumptions are not valid for the given subsystemNamesTableFile,
 * then this method will always return false.
 *
 * @param subsystemNameSpec The specification of the named subsystem
 *        configuration of interest.
 * @param idList The list of database IDs for the named configuration
 *        and its ancestors [output parameter].
 *        (If the status returned by this method is false, then the
 *        values returned in this parameter are undefined.)
 * @param subsystemNamesTableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 *
 * @returns true if the lookup of the ID of the specified named
 *        configuration and its ancestors worked, false if not.
 */
bool NamedConfigUtils::
getSubsystemConfigIdAncestry(const SubsystemNameSpec& subsystemNameSpec,
                             std::vector<int64_t>& idList,
                             const std::string& subsystemNamesTableFile)
{
    int64_t subsystemConfigId;

    boost::shared_ptr<NOVADB::Table> dbt;
    if (subsystemNameSpec.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }

    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    NamedConfigSelectionCriteria selections;
    selections[NAME_COLUMN_NAME] = subsystemNameSpec.getConfigName();
    selections[DCMOPMODE_COLUMN_NAME] = subsystemNameSpec.getDCMOpMode();
    selections[DETECTORID_COLUMN_NAME] = subsystemNameSpec.getDetectorId();
    selections[SUBSYSTEMNAME_COLUMN_NAME]=subsystemNameSpec.getSubsystemName();
    selections[CONFIGTYPE_COLUMN_NAME] = subsystemNameSpec.getConfigType();
    if (subsystemNameSpec.hasTimestamp()) {
        selections[INSERTTIME_COLUMN_NAME] = subsystemNameSpec.getTimestamp();
    }
    selections[ISOBSOLETE_COLUMN_NAME] = "false";
    selections[ISCHANGING_COLUMN_NAME] = "false";
    selections[ISCOMPLETE_COLUMN_NAME] = "true";

    if (! _getId(dbt, selections, subsystemConfigId)) {
        return false;
    }

    idList.push_back(subsystemConfigId);
    uint32_t previousSize = idList.size() - 1;

    while (idList.size() > previousSize) {
        previousSize = idList.size();
        if (! _findNextAncestor(dbt, idList)) {
            return false;
        }
    }
    return true;
}

/**
 * Fetches the database ID for the named subsystem configuration that
 * matches the input global name specification.
 *
 * NOTE that this method assumes certain column names for the database
 * table that contains the configuration names.  If any of these
 * assumptions are not valid for the given globalNamesTableFile,
 * then this method will always return false.
 *
 * @param globalNameSpec The specification of the named global
 *        configuration of interest.
 * @param subsystemNameSpec The specification of the subsystem
 *        of interest (only the subsystemName and configType are used).
 * @param subsystemConfigId The database ID of the named subsystem
 *        configuration [output parameter].
 *        (If the status returned by this method is false, then the value
 *        returned in this parameter is undefined.)
 * @param globalNamesTableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 * @param subsystemNamesTableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 *
 * @returns true if the lookup of the ID of the specified named
 *        configuration worked, false if not.
 */
bool NamedConfigUtils::
getSubsystemConfigIdFromGlobalName(const GlobalNameSpec& globalNameSpec,
                                   const std::string& subsystemName,
                                   const std::string& subsystemConfigType,
                                   int64_t& subsystemConfigId,
                                   const std::string& globalNamesTableFile,
                                   const std::string& subsystemNamesTableFile)
{
    std::vector<int64_t> globalIdList;
    if (! getGlobalConfigIdAncestry(globalNameSpec, globalIdList,
                                    globalNamesTableFile)) {
        return false;
    }

    boost::shared_ptr<NOVADB::Table> dbt;
    if (globalNameSpec.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }
    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();
    for (uint32_t cfgIdx = 0; cfgIdx < globalIdList.size(); ++cfgIdx) {
        dbt->Clear();
        dbt->SetValidityRange(GLOBALCFGID_COLUMN_NAME, globalIdList[cfgIdx]);
        dbt->SetValidityRange(SUBSYSTEMNAME_COLUMN_NAME, subsystemName);
        dbt->SetValidityRange(CONFIGTYPE_COLUMN_NAME, subsystemConfigType);
        dbt->SetValidityRange(ISOBSOLETE_COLUMN_NAME, false);
        dbt->SetValidityRange(ISCHANGING_COLUMN_NAME, false);
        dbt->SetValidityRange(ISCOMPLETE_COLUMN_NAME, true);
        if (! dbt->LoadFromDB()) {return false;}
        if (dbt->NRow() == 1) {
            NOVADB::Row* rowPtr = dbt->GetRow(0);
            NOVADB::Column* colPtr = rowPtr->Col(ID_COLUMN_NAME);
            return (colPtr->Get(subsystemConfigId));
        }
    }

    return false;
}

bool NamedConfigUtils::
getSubsystemConfigIdFromGlobalId(const IDSpec& idSpec,
                                 const std::string& subsystemName,
                                 const std::string& subsystemConfigType,
                                 int64_t& subsystemConfigId,
                                 const std::string& globalNamesTableFile,
                                 const std::string& subsystemNamesTableFile)
{
    std::vector<int64_t> globalIdList;
    if (! getConfigIdAncestry(idSpec, globalIdList, globalNamesTableFile)) {
        return false;
    }

    subsystemConfigId = 0;
    boost::shared_ptr<NOVADB::Table> dbt;
    if (idSpec.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }

    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();
    for (uint32_t cfgIdx = 0; cfgIdx < globalIdList.size(); ++cfgIdx) {
        dbt->Clear();
        dbt->SetValidityRange(GLOBALCFGID_COLUMN_NAME, globalIdList[cfgIdx]);
        dbt->SetValidityRange(SUBSYSTEMNAME_COLUMN_NAME, subsystemName);
        dbt->SetValidityRange(CONFIGTYPE_COLUMN_NAME, subsystemConfigType);
        if (! dbt->LoadFromDB()) {return false;}
        if (dbt->NRow() == 1) {
            NOVADB::Row* rowPtr = dbt->GetRow(0);
            NOVADB::Column* colPtr = rowPtr->Col(ID_COLUMN_NAME);
            if (! colPtr->Get(subsystemConfigId)) {return false;}
            else {break;}
        }
    }

    return (subsystemConfigId != 0);
}

bool NamedConfigUtils::
getSubsystemConfigIdAncestryFromGlobalId(const IDSpec& idSpec,
                                         const std::string& subsystemName,
                                         const std::string& subsystemConfigType,
                                         std::vector<int64_t>& idList,
                                         const std::string&
                                         globalNamesTableFile,
                                         const std::string&
                                         subsystemNamesTableFile)
{
    std::string gntf = globalNamesTableFile;
    if (idSpec.isSlowControlsConfiguration() &&
        globalNamesTableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        gntf = DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE;
    }
    std::string sntf = subsystemNamesTableFile;
    if (idSpec.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        sntf = DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE;
    }

    std::vector<int64_t> globalIdList;
    if (! getConfigIdAncestry(idSpec, globalIdList, gntf)) {
        return false;
    }

    int64_t subsystemConfigId = 0;
    boost::shared_ptr<NOVADB::Table> dbt;
    dbt.reset(new NOVADB::Table(sntf.c_str()));
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }

    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    dbt->GetConnection();
    for (uint32_t cfgIdx = 0; cfgIdx < globalIdList.size(); ++cfgIdx) {
        dbt->Clear();
        dbt->SetValidityRange(GLOBALCFGID_COLUMN_NAME, globalIdList[cfgIdx]);
        dbt->SetValidityRange(SUBSYSTEMNAME_COLUMN_NAME, subsystemName);
        dbt->SetValidityRange(CONFIGTYPE_COLUMN_NAME, subsystemConfigType);
        if (! dbt->LoadFromDB()) {return false;}
        if (dbt->NRow() == 1) {
            NOVADB::Row* rowPtr = dbt->GetRow(0);
            NOVADB::Column* colPtr = rowPtr->Col(ID_COLUMN_NAME);
            if (! colPtr->Get(subsystemConfigId)) {return false;}
            else {break;}
        }
    }
    if (subsystemConfigId == 0) {return false;}

    IDSpec subsystemIdSpec(subsystemConfigId, IDSpec::SUBSYSTEM,
                           (idSpec.isSlowControlsConfiguration() ?
                            IDSpec::SLOWCONTROLS : IDSpec::DAQ));
    return getConfigIdAncestry(subsystemIdSpec, idList, sntf);
}

bool NamedConfigUtils::
getSubsystemConfigIdsFromGlobalId(const IDSpec& idSpec,
                                  std::vector<int64_t>& idList,
                                  std::vector<SubsystemNameSpec>&
                                  subsystemNameSpecList,
                                  const std::string& globalNamesTableFile,
                                  const std::string& subsystemNamesTableFile)
{
    std::vector<int64_t> globalIdList;
    if (! getConfigIdAncestry(idSpec, globalIdList, globalNamesTableFile)) {
        return false;
    }

    boost::shared_ptr<NOVADB::Table> dbt;
    if (idSpec.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }
    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();
    for (uint32_t cfgIdx = 0; cfgIdx < globalIdList.size(); ++cfgIdx) {
        dbt->Clear();
        dbt->SetValidityRange(GLOBALCFGID_COLUMN_NAME, globalIdList[cfgIdx]);
        if (dbt->LoadFromDB()) {
            for (int iRow = 0; iRow < dbt->NRow(); ++iRow) {
                NOVADB::Row* rowPtr = dbt->GetRow(iRow);
                NOVADB::Column* colPtr;

                std::string subsystemName;
                colPtr = rowPtr->Col(SUBSYSTEMNAME_COLUMN_NAME);
                if (! colPtr->Get(subsystemName)) {return false;}

                std::string configType;
                colPtr = rowPtr->Col(CONFIGTYPE_COLUMN_NAME);
                if (! colPtr->Get(configType)) {return false;}

                bool alreadyFound = false;
                for (uint32_t sdx=0; sdx<subsystemNameSpecList.size(); ++sdx) {
                    if (subsystemName ==
                        subsystemNameSpecList[sdx].getSubsystemName() &&
                        configType ==
                        subsystemNameSpecList[sdx].getConfigType()) {
                        alreadyFound = true;
                        break;
                    }
                }
                if (alreadyFound) {continue;}

                int64_t subsystemConfigId = 0;
                colPtr = rowPtr->Col(ID_COLUMN_NAME);
                if (! colPtr->Get(subsystemConfigId)) {return false;}

                std::string configName;
                colPtr = rowPtr->Col(NAME_COLUMN_NAME);
                if (! colPtr->Get(configName)) {return false;}

                std::string dcmOpMode;
                colPtr = rowPtr->Col(DCMOPMODE_COLUMN_NAME);
                if (! colPtr->Get(dcmOpMode)) {return false;}

                std::string detectorName;
                colPtr = rowPtr->Col(DETECTORID_COLUMN_NAME);
                if (! colPtr->Get(detectorName)) {return false;}

                SubsystemNameSpec tmpNameSpec(configName, dcmOpMode,
                                              detectorName, subsystemName,
                                              configType);
                subsystemNameSpecList.push_back(tmpNameSpec);
                idList.push_back(subsystemConfigId);
            }
        }
    }

    return true;
}

std::vector<NamedConfigUtils::NamedConfigDetailsPtr> NamedConfigUtils::
getSubsystemConfigsFromGlobalId(const IDSpec& idSpec,
                                const std::string& globalNamesTableFile,
                                const std::string& subsystemNamesTableFile)
{
    std::vector<NamedConfigDetailsPtr> emptyList;
    std::vector<NamedConfigDetailsPtr> detailsList;

    std::vector<int64_t> globalIdList;
    if (! getConfigIdAncestry(idSpec, globalIdList, globalNamesTableFile)) {
        return emptyList;
    }

    boost::shared_ptr<NOVADB::Table> dbt;
    if (idSpec.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }
    std::string detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();
    for (uint32_t cfgIdx = 0; cfgIdx < globalIdList.size(); ++cfgIdx) {
        dbt->Clear();
        dbt->SetValidityRange(GLOBALCFGID_COLUMN_NAME, globalIdList[cfgIdx]);
        if (! dbt->LoadFromDB()) {
            return emptyList;
        }
        for (int iRow = 0; iRow < dbt->NRow(); ++iRow) {
            NamedConfigDetailsPtr detailsPtr(new NamedConfigDetails());
            std::string columnValue;
            NOVADB::Row* rowPtr = dbt->GetRow(iRow);
            for (int idx = 0; idx < dbt->NCol(); ++idx) {
                NOVADB::Column& colRef = rowPtr->Col(idx);
                if (colRef.IsNull()) {
                    (*detailsPtr)[colRef.Name()] = "";
                }
                else if (colRef.Get(columnValue)) {
                    (*detailsPtr)[colRef.Name()] = columnValue;
                }
                else {
                    return emptyList;
                }
            }

            bool alreadyFound = false;
            for (uint32_t ddx = 0; ddx < detailsList.size(); ++ddx) {
                std::string sname1, sname2, ctype1, ctype2;
                if (getFieldValue(detailsPtr,
                                  SUBSYSTEMNAME_COLUMN_NAME, sname1) &&
                    getFieldValue(detailsList[ddx],
                                  SUBSYSTEMNAME_COLUMN_NAME, sname2) &&
                    getFieldValue(detailsPtr,
                                  CONFIGTYPE_COLUMN_NAME, ctype1) &&
                    getFieldValue(detailsList[ddx],
                                  CONFIGTYPE_COLUMN_NAME, ctype2) &&
                    sname1 == sname2 && ctype1 == ctype2) {
                    alreadyFound = true;
                    break;
                }
            }
            if (! alreadyFound) {
                detailsList.push_back(detailsPtr);
            }
        }
    }

    return detailsList;
}

NamedConfigUtils::NamedConfigDetailsPtr NamedConfigUtils::
getNewestGlobalConfigFromSubsystemName(const SubsystemNameSpec& subsystemNameSpec,
                                       const std::string& subsystemNamesTableFile,
                                       const std::string& globalNamesTableFile)
{
    NamedConfigDetailsPtr detailsPtr;

    // step 1: look up the subsystem configuration
    boost::shared_ptr<NOVADB::Table> dbt;
    if (subsystemNameSpec.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    NamedConfigSelectionCriteria selections;
    selections[NAME_COLUMN_NAME] = subsystemNameSpec.getConfigName();
    selections[DCMOPMODE_COLUMN_NAME] = subsystemNameSpec.getDCMOpMode();
    selections[DETECTORID_COLUMN_NAME] = subsystemNameSpec.getDetectorId();
    selections[SUBSYSTEMNAME_COLUMN_NAME]=subsystemNameSpec.getSubsystemName();
    selections[CONFIGTYPE_COLUMN_NAME] = subsystemNameSpec.getConfigType();
    if (subsystemNameSpec.hasTimestamp()) {
        selections[INSERTTIME_COLUMN_NAME] = subsystemNameSpec.getTimestamp();
    }
    selections[ISOBSOLETE_COLUMN_NAME] = "false";
    selections[ISCHANGING_COLUMN_NAME] = "false";
    selections[ISCOMPLETE_COLUMN_NAME] = "true";

    NamedConfigDetailsPtr subsysDetails = _getDetails(dbt, selections);

    // step 2: pull out the associated global CFG ID
    int64_t globalCfgId = 0;
    std::string tmpString;
    if (! getFieldValue(subsysDetails, GLOBALCFGID_COLUMN_NAME,
                        tmpString)) {return detailsPtr;}
    try {
        globalCfgId = boost::lexical_cast<int64_t>(tmpString);
    }
    catch (...) {return detailsPtr;}

    // step 3: fetch the details for the latest global CFG ID
    IDSpec globalIdSpec(globalCfgId, IDSpec::GLOBAL,
                        (subsystemNameSpec.isSlowControlsConfiguration() ?
                         IDSpec::SLOWCONTROLS : IDSpec::DAQ));
    return getNewestGlobalConfigFromGlobalId(globalIdSpec,
                                             globalNamesTableFile);
}

NamedConfigUtils::NamedConfigDetailsPtr NamedConfigUtils::
getNewestGlobalConfigFromSubsystemId(const IDSpec& idSpec,
                                     const std::string& subsystemNamesTableFile,
                                     const std::string& globalNamesTableFile)
{
    NamedConfigDetailsPtr detailsPtr;

    // step 1: look up the subsystem configuration
    boost::shared_ptr<NOVADB::Table> dbt;
    if (idSpec.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::
                  Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    NamedConfigSelectionCriteria selections;
    selections[ID_COLUMN_NAME] =
        boost::lexical_cast<std::string>(idSpec.getConfigId());

    NamedConfigDetailsPtr subsysDetails = _getDetails(dbt, selections);

    // step 2: pull out the associated global CFG ID
    int64_t globalCfgId = 0;
    std::string tmpString;
    if (! getFieldValue(subsysDetails, GLOBALCFGID_COLUMN_NAME,
                        tmpString)) {return detailsPtr;}
    try {
        globalCfgId = boost::lexical_cast<int64_t>(tmpString);
    }
    catch (...) {return detailsPtr;}

    // step 3: fetch the details for the latest global CFG ID
    IDSpec globalIdSpec(globalCfgId, IDSpec::GLOBAL,
                        (idSpec.isSlowControlsConfiguration() ?
                         IDSpec::SLOWCONTROLS : IDSpec::DAQ));
    return getNewestGlobalConfigFromGlobalId(globalIdSpec,
                                             globalNamesTableFile);
}

NamedConfigUtils::NamedConfigDetailsPtr NamedConfigUtils::
getNewestGlobalConfigFromGlobalId(const IDSpec& idSpec,
                                  const std::string& globalNamesTableFile)
{
    NamedConfigDetailsPtr detailsPtr;

    // step 1: build the list of child global IDs
    boost::shared_ptr<NOVADB::Table> dbt;
    if (idSpec.isSlowControlsConfiguration() &&
        globalNamesTableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::
                  Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(globalNamesTableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    std::vector<int64_t> globalCfgIdList;
    globalCfgIdList.push_back(idSpec.getConfigId());
    uint32_t previousSize = 0;

    while (globalCfgIdList.size() > previousSize) {
        previousSize = globalCfgIdList.size();
        if (! _findNextChild(dbt, globalCfgIdList)) {
            return detailsPtr;
        }
    }
    int64_t globalCfgId = globalCfgIdList[globalCfgIdList.size()-1];

    // step 2: fetch the details for the latest global CFG ID
    NamedConfigSelectionCriteria selections;
    selections[ID_COLUMN_NAME] = boost::lexical_cast<std::string>(globalCfgId);
    detailsPtr = _getDetails(dbt, selections);

    return detailsPtr;
}

/**
 * Fetches the details of the most recent global configuration that was
 * used by the specified partition.
 *
 * @returns an empty shared pointer if the lookup fails.
 */
NamedConfigUtils::GlobalNameSpecPtr NamedConfigUtils::
getNewestInstanceOfLatestUsedDAQGlobalConfig(const std::string& detectorName,
                                             const int& partitionNumber,
                                             const std::string& globalNamesTableFile)
{
    GlobalNameSpecPtr nameSpecPtr;
    boost::shared_ptr<NOVADB::Table> dbt;
    dbt.reset(new NOVADB::Table(globalNamesTableFile.c_str()));
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    else {
        dbUser = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBUSER");
         if (dbUser.length() > 0) {
            dbt->SetUser(dbUser);
        }
    }
    dbt->GetConnection();

    PGresult* resultSetPtr;
    std::string queryString="SELECT cfg.name,cfg.dcmopmode,cfg.detectorname ";
    queryString.append("FROM runs as rr ");
    queryString.append("JOIN " + dbt->Schema() + "." + dbt->Name() + " as cfg ");
    queryString.append("ON (rr.configid = cfg.id) ");
    queryString.append("WHERE rr.partition=");
    queryString.append(boost::lexical_cast<std::string>(partitionNumber));
    queryString.append(" AND rr.detector='");
    queryString.append(detectorName);
    queryString.append("' ORDER BY rr.run DESC LIMIT 1");

    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(queryString, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_TUPLES_OK) {
        PQclear(resultSetPtr);
        return nameSpecPtr;
    }

    int nRows = PQntuples(resultSetPtr);
    if (nRows != 1) {
        PQclear(resultSetPtr);
        return nameSpecPtr;
    }

    std::string name1 = PQgetvalue(resultSetPtr, 0, 0);
    std::string opmode1 = PQgetvalue(resultSetPtr, 0, 1);
    std::string detector1 = PQgetvalue(resultSetPtr, 0, 2);
    PQclear(resultSetPtr);

    GlobalNameSpec globalNameSpec(name1, opmode1, detector1);
    std::vector<NamedConfigDetailsPtr> detailsList =
        getGlobalConfigDetailAncestry(globalNameSpec, globalNamesTableFile);
    if (detailsList.size() == 0) {return nameSpecPtr;}

    std::string nameString;
    std::string opModeString;
    std::string detectorString;
    std::string timeString;
    if (! getFieldValue(detailsList[0], NAME_COLUMN_NAME, nameString) ||
        ! getFieldValue(detailsList[0], DCMOPMODE_COLUMN_NAME, opModeString) ||
        ! getFieldValue(detailsList[0], DETECTORNAME_COLUMN_NAME, detectorString) ||
        ! getFieldValue(detailsList[0], INSERTTIME_COLUMN_NAME, timeString)) {
        return nameSpecPtr;
    }
    nameSpecPtr.reset(new GlobalNameSpec(nameString, opModeString,
                                         detectorString, timeString));
    return nameSpecPtr;
}

/**
 * Fetches a list of recent global configurations for a specific detector.
 */
bool NamedConfigUtils::
getRecentGlobalConfigs(IDSpec::MAJOR_SYSTEM_TYPE daqOrSlowControls,
                       const std::string& detectorName,
                       int resultOffset, int resultLimit,
                       GlobalNameSpecList& resultList,
                       const std::string& globalNamesTableFile)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    if (daqOrSlowControls == IDSpec::SLOWCONTROLS &&
        globalNamesTableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::
                  Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(globalNamesTableFile.c_str()));
    }
    std::string detName = detectorName;
    if (!novadaq::cnv::DetInfo::DetNameOk(detName)) 
      detName = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->SetTableType(NOVADB::kDAQTable);

    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    else {
        dbUser = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBUSER");
         if (dbUser.length() > 0) {
            dbt->SetUser(dbUser);
        }
    }
    dbt->GetConnection();

    dbt->Clear();
    //    dbt->SetValidityRange(DETECTORNAME_COLUMN_NAME, detectorName);
    dbt->SetValidityRange(ISOBSOLETE_COLUMN_NAME, false);
    dbt->SetValidityRange(ISCHANGING_COLUMN_NAME, false);
    dbt->SetValidityRange(ISCOMPLETE_COLUMN_NAME, true);
    dbt->AddOrderColumn(INSERTTIME_COLUMN_NAME);
    dbt->SetOrderDesc();
    dbt->SetSelectLimit(resultLimit);
    dbt->SetSelectOffset(resultOffset);
    if (! dbt->LoadFromDB()) {return false;}

    for (int32_t idx = 0; idx < dbt->NRow(); ++idx) {
        NOVADB::Row* rowPtr = dbt->GetRow(idx);
        NOVADB::Column* colPtr;

        std::string configName;
        colPtr = rowPtr->Col(NAME_COLUMN_NAME);
        if (! colPtr->Get(configName)) {return false;}

        std::string dcmOpMode;
        colPtr = rowPtr->Col(DCMOPMODE_COLUMN_NAME);
        if (! colPtr->Get(dcmOpMode)) {return false;}

        std::string timestamp;
        colPtr = rowPtr->Col(INSERTTIME_COLUMN_NAME);
        if (! colPtr->Get(timestamp)) {return false;}

        GlobalNameSpec nameSpec(configName, dcmOpMode,
                                detectorName, timestamp);

        resultList.push_back(nameSpec);
    }

    return true;
}

/**
 * Fetches a list of the available global configuration names and DCM
 * operating modes for a specific detector.
 */
bool NamedConfigUtils::
getAvailableGlobalConfigs(IDSpec::MAJOR_SYSTEM_TYPE daqOrSlowControls,
                          const std::string& detectorName,
                          GlobalNameSpecMap& resultsTable,
                          int maxDetailCountPerSummary,
                          const std::string& globalNamesTableFile)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    if (daqOrSlowControls == IDSpec::SLOWCONTROLS &&
        globalNamesTableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::
                  Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(globalNamesTableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    else {
        dbUser = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBUSER");
        if (dbUser.length() > 0) {
            dbt->SetUser(dbUser);
        }
    }
    dbt->GetConnection();

    // Fetch the list of available Name and DCMOpMode combinations,
    // for all timestamps and taking into account all usageCounts.
    PGresult* resultSetPtr;
    std::string queryString = "SELECT ";
    queryString.append(NAME_COLUMN_NAME);
    queryString.append(", ");
    queryString.append(DCMOPMODE_COLUMN_NAME);
    queryString.append(", SUM(");
    queryString.append(USAGECOUNT_COLUMN_NAME);
    queryString.append("), MAX(");
    queryString.append(INSERTTIME_COLUMN_NAME);
    queryString.append(") FROM ");
    queryString.append(dbt->Schema() + "." + dbt->Name());
    queryString.append(" WHERE ");
    //    queryString.append(DETECTORNAME_COLUMN_NAME);
    //    queryString.append("='");
    //    queryString.append(detectorName);
    //    queryString.append("' AND ");
    queryString.append(ISOBSOLETE_COLUMN_NAME);
    queryString.append("=false AND ");
    queryString.append(ISCHANGING_COLUMN_NAME);
    queryString.append("=false AND ");
    queryString.append(ISCOMPLETE_COLUMN_NAME);
    queryString.append("=true GROUP BY ");
    queryString.append(NAME_COLUMN_NAME);
    queryString.append(", ");
    queryString.append(DCMOPMODE_COLUMN_NAME);
    queryString.append(" ORDER BY SUM(");
    queryString.append(USAGECOUNT_COLUMN_NAME);
    queryString.append(") DESC, MAX(");
    queryString.append(INSERTTIME_COLUMN_NAME);
    queryString.append(") DESC, ");
    queryString.append(NAME_COLUMN_NAME);
    queryString.append(" ASC");

    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(queryString, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_TUPLES_OK) {
        PQclear(resultSetPtr);

        return false;
    }
    std::vector<GlobalNameSpec> summaryList;
    int nRows = PQntuples(resultSetPtr);
    if (nRows > 0) {
        for (int idx = 0; idx < nRows; ++idx) {
            std::string nameString = PQgetvalue(resultSetPtr, idx, 0);
            std::string opModeString = PQgetvalue(resultSetPtr, idx, 1);
            std::string usageCountString = PQgetvalue(resultSetPtr, idx, 2);
            int usageCount = 0;
            try {
                usageCount = boost::lexical_cast<int>(usageCountString);
            }
            catch (...) {}

            GlobalNameSpec nameSpec(nameString, opModeString, detectorName);
            nameSpec.setUsageCount(usageCount);
            nameSpec.setSortIndex(idx);
            summaryList.push_back(nameSpec);
        }
    }
    PQclear(resultSetPtr);

    // Fetch the most recent N instances of each name and dcmOpMode.
    for (int cfgIdx = 0; cfgIdx < (int) summaryList.size(); ++cfgIdx) {
        std::string configName = summaryList[cfgIdx].getConfigName();
        std::string dcmOpMode = summaryList[cfgIdx].getDCMOpMode();

        dbt->Clear();
        dbt->SetValidityRange(NAME_COLUMN_NAME, configName);
        dbt->SetValidityRange(DCMOPMODE_COLUMN_NAME, dcmOpMode);
	//        dbt->SetValidityRange(DETECTORNAME_COLUMN_NAME, detectorName);
        dbt->SetValidityRange(ISOBSOLETE_COLUMN_NAME, false);
        dbt->SetValidityRange(ISCHANGING_COLUMN_NAME, false);
        dbt->SetValidityRange(ISCOMPLETE_COLUMN_NAME, true);
        dbt->AddOrderColumn(INSERTTIME_COLUMN_NAME);
        dbt->SetOrderDesc();
        dbt->SetSelectLimit(maxDetailCountPerSummary);
        if (! dbt->LoadFromDB()) {return false;}

        std::vector<GlobalNameSpec> detailsList;
        for (int32_t idx = 0; idx < dbt->NRow(); ++idx) {
            NOVADB::Row* rowPtr = dbt->GetRow(idx);
            NOVADB::Column* colPtr;

            std::string timestamp;
            colPtr = rowPtr->Col(INSERTTIME_COLUMN_NAME);
            if (! colPtr->Get(timestamp)) {return false;}

            int usageCount;
            colPtr = rowPtr->Col(USAGECOUNT_COLUMN_NAME);
            if (colPtr->IsNull()) {usageCount = 0;}
            else if (! colPtr->Get(usageCount)) {return false;}

            GlobalNameSpec nameSpec(configName, dcmOpMode,
                                    detectorName, timestamp);
            nameSpec.setUsageCount(usageCount);
            detailsList.push_back(nameSpec);
        }

        resultsTable[summaryList[cfgIdx]] = detailsList;
    }

    return true;
}

/**
 * Fetches a list of the available subsystem configuration names and DCM
 * operating modes for a specific subsystem, configuration type, and
 * detector ID.
 *
 * NOTE that this method assumes certain column names for the database
 * table that contains the configuration names.  If any of these
 * assumptions are not valid for the given globalNamesTableFile,
 * then this method will always return false.
 *
 * @param subsystemName The name of the subsystem of interest.
 * @param configType The configuration type of interest.
 * @param detectorName The detector of interest (e.g. NDOS or FarDet).
 * @param subsystemConfigList The resulting list of subsystem configurations
 *        [output parameter].
 *        (If the status returned by this method is false, then the value
 *        returned in this parameter is undefined.)
 * @param subsystemNamesTableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 *
 * @returns true if the lookup of the available subsystem configuration
 *        names worked, false if not.
 */
bool NamedConfigUtils::
getAvailableSubsystemConfigs(IDSpec::MAJOR_SYSTEM_TYPE daqOrSlowControls,
                             const std::string& detectorName,
                             const std::string& subsystemName,
                             const std::string& configType,
                             SubsystemNameSpecMap& resultsTable,
                             int maxDetailCountPerSummary,
                             const std::string& subsystemNamesTableFile)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    if (daqOrSlowControls == IDSpec::SLOWCONTROLS &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::
                  Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    else {
        dbUser = novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBUSER");
        if (dbUser.length() > 0) {
            dbt->SetUser(dbUser);
        }
    }
    dbt->GetConnection();

    // Fetch the list of available Name and DCMOpMode combinations,
    // for all timestamps and taking into account all usageCounts.
    PGresult* resultSetPtr;
    std::string queryString = "SELECT ";
    queryString.append(NAME_COLUMN_NAME);
    queryString.append(", ");
    queryString.append(DCMOPMODE_COLUMN_NAME);
    queryString.append(", SUM(");
    queryString.append(USAGECOUNT_COLUMN_NAME);
    queryString.append("), MAX(");
    queryString.append(INSERTTIME_COLUMN_NAME);
    queryString.append(") FROM ");
    queryString.append(dbt->Schema() + "." + dbt->Name());
    queryString.append(" WHERE ");
    //    queryString.append(DETECTORNAME_COLUMN_NAME);
    //    queryString.append("='");
    //    queryString.append(detectorName);
    //    queryString.append("' AND ");
    queryString.append(SUBSYSTEMNAME_COLUMN_NAME);
    queryString.append("='");
    queryString.append(subsystemName);
    queryString.append("' AND ");
    queryString.append(CONFIGTYPE_COLUMN_NAME);
    queryString.append("='");
    queryString.append(configType);
    queryString.append("' AND ");
    queryString.append(ISOBSOLETE_COLUMN_NAME);
    queryString.append("=false AND ");
    queryString.append(ISCHANGING_COLUMN_NAME);
    queryString.append("=false AND ");
    queryString.append(ISCOMPLETE_COLUMN_NAME);
    queryString.append("=true GROUP BY ");
    queryString.append(NAME_COLUMN_NAME);
    queryString.append(", ");
    queryString.append(DCMOPMODE_COLUMN_NAME);
    queryString.append(" ORDER BY SUM(");
    queryString.append(USAGECOUNT_COLUMN_NAME);
    queryString.append(") DESC, MAX(");
    queryString.append(INSERTTIME_COLUMN_NAME);
    queryString.append(") DESC, ");
    queryString.append(NAME_COLUMN_NAME);
    queryString.append(" ASC");

    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(queryString, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_TUPLES_OK) {
        PQclear(resultSetPtr);

        return false;
    }
    std::vector<SubsystemNameSpec> summaryList;
    int nRows = PQntuples(resultSetPtr);
    if (nRows > 0) {
        for (int idx = 0; idx < nRows; ++idx) {
            std::string nameString = PQgetvalue(resultSetPtr, idx, 0);
            std::string opModeString = PQgetvalue(resultSetPtr, idx, 1);
            std::string usageCountString = PQgetvalue(resultSetPtr, idx, 2);
            int usageCount = 0;
            try {
                usageCount = boost::lexical_cast<int>(usageCountString);
            }
            catch (...) {}

            SubsystemNameSpec nameSpec(nameString, opModeString, detectorName,
                                       subsystemName, configType);
            nameSpec.setUsageCount(usageCount);
            nameSpec.setSortIndex(idx);
            summaryList.push_back(nameSpec);
        }
    }
    PQclear(resultSetPtr);

    // Fetch the most recent N instances of each name and dcmOpMode.
    for (int cfgIdx = 0; cfgIdx < (int) summaryList.size(); ++cfgIdx) {
        std::string configName = summaryList[cfgIdx].getConfigName();
        std::string dcmOpMode = summaryList[cfgIdx].getDCMOpMode();

        dbt->Clear();
        dbt->SetValidityRange(NAME_COLUMN_NAME, configName);
        dbt->SetValidityRange(DCMOPMODE_COLUMN_NAME, dcmOpMode);
        dbt->SetValidityRange(DETECTORNAME_COLUMN_NAME, detectorName);
        dbt->SetValidityRange(SUBSYSTEMNAME_COLUMN_NAME, subsystemName);
        dbt->SetValidityRange(CONFIGTYPE_COLUMN_NAME, configType);
        dbt->SetValidityRange(ISOBSOLETE_COLUMN_NAME, false);
        dbt->SetValidityRange(ISCHANGING_COLUMN_NAME, false);
        dbt->SetValidityRange(ISCOMPLETE_COLUMN_NAME, true);
        dbt->AddOrderColumn(INSERTTIME_COLUMN_NAME);
        dbt->SetOrderDesc();
        dbt->SetSelectLimit(maxDetailCountPerSummary);
        if (! dbt->LoadFromDB()) {return false;}

        std::vector<SubsystemNameSpec> detailsList;
        for (int32_t idx = 0; idx < dbt->NRow(); ++idx) {
            NOVADB::Row* rowPtr = dbt->GetRow(idx);
            NOVADB::Column* colPtr;

            std::string timestamp;
            colPtr = rowPtr->Col(INSERTTIME_COLUMN_NAME);
            if (! colPtr->Get(timestamp)) {return false;}

            int usageCount;
            colPtr = rowPtr->Col(USAGECOUNT_COLUMN_NAME);
            if (colPtr->IsNull()) {usageCount = 0;}
            else if (! colPtr->Get(usageCount)) {return false;}

            SubsystemNameSpec nameSpec(configName, dcmOpMode, detectorName,
                                       subsystemName, configType, timestamp);
            nameSpec.setUsageCount(usageCount);
            detailsList.push_back(nameSpec);
        }

        resultsTable[summaryList[cfgIdx]] = detailsList;
    }

    return true;
}

novadaq::Status NamedConfigUtils::
copyGlobalConfig(const IDSpec& sourceId, bool makeParentChildLink,
                 const SimpleParameterSet& valuesToOverride,
                 std::string baselineTimestamp,
                 bool overrideIntermediateUpdate,
                 int64_t& resultId, const std::string& globalNamesTableFile)
{
    novadaq::Status resultStatus;
    PGresult* resultSetPtr;
    boost::shared_ptr<NOVADB::Table> dbt;
    if (sourceId.isSlowControlsConfiguration() &&
        globalNamesTableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::
                  Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(globalNamesTableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    else {
        dbUser = novadaq::EnvVarCache::getInstance().getEnvVar("USER");
    }
    dbt->GetConnection();

    // look up the parameters associated with the source configuration
    NamedConfigSelectionCriteria selections;
    selections[ID_COLUMN_NAME] =
        boost::lexical_cast<std::string>(sourceId.getConfigId());
    NamedConfigDetailsPtr sourceCfgDetails = _getDetails(dbt, selections);
    if (sourceCfgDetails.get() == 0) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "The lookup of the input global ";
        resultStatus.message.append("configuration ID failed.");
        return resultStatus;
    }
    std::string sourceConfigName;
    if (! getFieldValue(sourceCfgDetails, NAME_COLUMN_NAME,
                        sourceConfigName)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the configuration name ";
        resultStatus.message.append("for the source configuration.");
        return resultStatus;
    }
    std::string sourceDcmOpMode;
    if (! getFieldValue(sourceCfgDetails, DCMOPMODE_COLUMN_NAME,
                        sourceDcmOpMode)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the DCM operating mode ";
        resultStatus.message.append("for the source configuration.");
        return resultStatus;
    }
    std::string sourceDetectorName;
    if (! getFieldValue(sourceCfgDetails, DETECTORNAME_COLUMN_NAME,
                        sourceDetectorName)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the detector name ";
        resultStatus.message.append("for the source configuration.");
        return resultStatus;
    }
    std::string sourceTimestamp;
    if (! getFieldValue(sourceCfgDetails, INSERTTIME_COLUMN_NAME,
                        sourceTimestamp)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the timestamp for the ";
        resultStatus.message.append("source configuration.");
        return resultStatus;
    }

    // determine the values of the parameters that we'll use to identify
    // the target configuration
    bool targetAndSourceHaveSameKey = true;
    std::string targetConfigName;
    if (valuesToOverride.get(NAME_COLUMN_NAME, targetConfigName)) {
        if (targetConfigName != sourceConfigName) {
            targetAndSourceHaveSameKey = false;
        }
    }
    else {
        targetConfigName = sourceConfigName;
    }
    std::string targetDcmOpMode;
    if (valuesToOverride.get(DCMOPMODE_COLUMN_NAME, targetDcmOpMode)) {
        if (targetDcmOpMode != sourceDcmOpMode) {
            targetAndSourceHaveSameKey = false;
        }
    }
    else {
        targetDcmOpMode = sourceDcmOpMode;
    }
    std::string targetDetectorName;
    if (valuesToOverride.get(DETECTORNAME_COLUMN_NAME, targetDetectorName)) {
        if (targetDetectorName != sourceDetectorName) {
            targetAndSourceHaveSameKey = false;
        }
    }
    else {
        targetDetectorName = sourceDetectorName;
    }

    // lock the table
    std::string dbCmd = "BEGIN";
    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(dbCmd, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_COMMAND_OK) {
        PQclear(resultSetPtr);

        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to begin the database transaction.";
        return resultStatus;
    }
    PQclear(resultSetPtr);
    dbCmd = "LOCK TABLE ";
    dbCmd.append(dbt->Schema() + "." + dbt->Name());
    dbCmd.append(" IN ACCESS EXCLUSIVE MODE");
    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(dbCmd, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_COMMAND_OK) {
        PQclear(resultSetPtr);

        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to lock DB table ";
        resultStatus.message.append(dbt->Schema() + "." + dbt->Name());
        resultStatus.message.append(".");
        return resultStatus;
    }
    PQclear(resultSetPtr);

    // check if the latest instance of the target configuration is changing
    std::string queryString = "SELECT ";
    queryString.append(ISCHANGING_COLUMN_NAME);
    queryString.append(", ");
    queryString.append(ID_COLUMN_NAME);
    queryString.append(" FROM ");
    queryString.append(dbt->Schema() + "." + dbt->Name());
    queryString.append(" WHERE ");
    queryString.append(NAME_COLUMN_NAME);
    queryString.append("='");
    queryString.append(targetConfigName);
    queryString.append("' AND ");
    queryString.append(DCMOPMODE_COLUMN_NAME);
    queryString.append("='");
    queryString.append(targetDcmOpMode);
    queryString.append("' AND ");
    queryString.append(DETECTORNAME_COLUMN_NAME);
    queryString.append("='");
    queryString.append(targetDetectorName);
    queryString.append("' ORDER BY ");
    queryString.append(ID_COLUMN_NAME);
    queryString.append(" DESC LIMIT 1");

    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(queryString, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_TUPLES_OK) {
        PQclear(resultSetPtr);

        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to check if the latest instance ";
        resultStatus.message.append("of the target configuration ");
        resultStatus.message.append("is changing.");
        return resultStatus;
    }
    if (PQntuples(resultSetPtr) > 0) {
        std::string resultString = PQgetvalue(resultSetPtr, 0, 0);
        boost::to_lower(resultString);
        if (resultString == "t") {
            resultStatus.code = COPY_FOUND_UPDATE_IN_PROGRESS;
            resultStatus.message = "Concurrent changes are in progress ";
            resultStatus.message.append("(globalConfigId = ");
            resultStatus.message.append(PQgetvalue(resultSetPtr, 0, 1));
            resultStatus.message.append(").");

            PQclear(resultSetPtr);

            dbCmd = "END";
            dbt->ExecuteSQL(dbCmd, resultSetPtr);

            return resultStatus;
        }
    }
    PQclear(resultSetPtr);

    // check if there are intermediate changes to the target configuration
    if (! overrideIntermediateUpdate) {
        boost::trim(baselineTimestamp);
        if (baselineTimestamp.length() == 0) {
            if (targetAndSourceHaveSameKey) {
                baselineTimestamp = sourceTimestamp;
            }
            else {
                baselineTimestamp = NOVADB::Util::GetFarPastTimeAsString();
            }
        }

        queryString = "SELECT ";
        queryString.append(ID_COLUMN_NAME);
        queryString.append(" FROM ");
        queryString.append(dbt->Schema() + "." + dbt->Name());
        queryString.append(" WHERE ");
        queryString.append(NAME_COLUMN_NAME);
        queryString.append("='");
        queryString.append(targetConfigName);
        queryString.append("' AND ");
        queryString.append(DCMOPMODE_COLUMN_NAME);
        queryString.append("='");
        queryString.append(targetDcmOpMode);
        queryString.append("' AND ");
        queryString.append(DETECTORNAME_COLUMN_NAME);
        queryString.append("='");
        queryString.append(targetDetectorName);
        queryString.append("' AND ");
        queryString.append(INSERTTIME_COLUMN_NAME);
        queryString.append(">'");
        queryString.append(baselineTimestamp);
        queryString.append("' ORDER BY ");
        queryString.append(ID_COLUMN_NAME);
        queryString.append(" DESC");

        resultSetPtr = 0;
        if (! dbt->ExecuteSQL(queryString, resultSetPtr) ||
            PQresultStatus(resultSetPtr) != PGRES_TUPLES_OK) {
            PQclear(resultSetPtr);

            resultStatus.code = COPY_FAILED;
            resultStatus.message = "Unable to check if there are ";
            resultStatus.message.append("intermediate changes in the ");
            resultStatus.message.append("target configuration.");
            return resultStatus;
        }
        if (PQntuples(resultSetPtr) > 0) {
            resultStatus.code = COPY_FOUND_INTERMEDIATE_UPDATE;
            resultStatus.message = "Found intermediate changes in the ";
            resultStatus.message.append("target configuration ");
            resultStatus.message.append("(e.g. globalConfigId = ");
            resultStatus.message.append(PQgetvalue(resultSetPtr, 0, 0));
            resultStatus.message.append(").");

            PQclear(resultSetPtr);

            dbCmd = "END";
            dbt->ExecuteSQL(dbCmd, resultSetPtr);

            return resultStatus;
        }
        PQclear(resultSetPtr);
    }

    // add the new target row
    const NOVADB::Column* colPtr;
    std::string columnNamesClause;
    std::string valuesClause;
    std::vector<std::string> columnNames = dbt->GetColNames();
    for (uint32_t idx = 0; idx < columnNames.size(); ++idx) {

        std::string columnName = boost::to_lower_copy(columnNames[idx]);
        if (columnName == NamedConfigUtils::ID_COLUMN_NAME) {continue;}
        if (columnName == NamedConfigUtils::USAGECOUNT_COLUMN_NAME) {continue;}
        if (columnName == NamedConfigUtils::UPDATETIME_COLUMN_NAME) {continue;}
        if (columnName == NamedConfigUtils::UPDATEUSER_COLUMN_NAME) {continue;}

        if (columnName == NamedConfigUtils::ISCHANGING_COLUMN_NAME) {
            if (columnNamesClause.length() > 0) {
                columnNamesClause.append(",");
                valuesClause.append(",");
            }
            columnNamesClause.append(columnName);
            valuesClause.append("true");
            continue;
        }

        if (columnName == NamedConfigUtils::ISCOMPLETE_COLUMN_NAME) {
            if (columnNamesClause.length() > 0) {
                columnNamesClause.append(",");
                valuesClause.append(",");
            }
            columnNamesClause.append(columnName);
            valuesClause.append("false");
            continue;
        }

        if (columnName == NamedConfigUtils::PARENTID_COLUMN_NAME) {
            if (makeParentChildLink) {
                if (columnNamesClause.length() > 0) {
                    columnNamesClause.append(",");
                    valuesClause.append(",");
                }
                columnNamesClause.append(columnName);
                valuesClause.append(boost::lexical_cast<std::string>
                                    (sourceId.getConfigId()));
            }
            continue;
        }

        if (columnName == NamedConfigUtils::INSERTTIME_COLUMN_NAME) {
            if (columnNamesClause.length() > 0) {
                columnNamesClause.append(",");
                valuesClause.append(",");
            }
            columnNamesClause.append(columnName);
            valuesClause.append("'");
            valuesClause.append(NOVADB::Util::GetCurrentTimeAsString());
            valuesClause.append("'");
            continue;
        }

        if (columnName == NamedConfigUtils::INSERTUSER_COLUMN_NAME) {
            if (columnNamesClause.length() > 0) {
                columnNamesClause.append(",");
                valuesClause.append(",");
            }
            columnNamesClause.append(columnName);
            valuesClause.append("'");
            valuesClause.append(dbUser);
            valuesClause.append("'");
            continue;
        }

        std::string value;
        if (! valuesToOverride.get(columnName, value) && 
            ! getFieldValue(sourceCfgDetails, columnName, value)) {
            resultStatus.code = COPY_FAILED;
            resultStatus.message = "The lookup of the value for the ";
            resultStatus.message.append(columnName);
            resultStatus.message.append(" column failed.");
            return resultStatus;
        }

        colPtr = dbt->GetCol(columnName);
        std::string columnType = "string";
        if (colPtr != 0) {
            columnType = colPtr->Type();
        }

        if (columnNamesClause.length() > 0) {
            columnNamesClause.append(",");
            valuesClause.append(",");
        }
        columnNamesClause.append(columnName);
        if (columnType == "string" || columnType == "timestamp") {
            valuesClause.append("'");
            valuesClause.append(value);
            valuesClause.append("'");
        }
        else if (columnType == "bool") {
            try {
                int intValue = boost::lexical_cast<int>(value);
                if (intValue) {
                    valuesClause.append("true");
                }
                else {
                    valuesClause.append("false");
                }
            }
            catch (...) {
                valuesClause.append(value);
            }
        }
        else {
            valuesClause.append(value);
        }
    }

    std::string insertString = "INSERT INTO ";
    insertString.append(dbt->Schema() + "." + dbt->Name());
    insertString.append(" (");
    insertString.append(columnNamesClause);
    insertString.append(") VALUES (");
    insertString.append(valuesClause);
    insertString.append(")");

    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(insertString, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_COMMAND_OK) {
        PQclear(resultSetPtr);

        resultStatus.code = COPY_FAILED;
        resultStatus.message = "The writing of the new global configuration ";
        resultStatus.message.append("into the database failed.");
        return resultStatus;
    }
    PQclear(resultSetPtr);
    if (! dbt->GetCurrSeqVal(NamedConfigUtils::ID_COLUMN_NAME, resultId)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "The lookup of the new global configuration ";
        resultStatus.message.append("ID failed.");
        return resultStatus;
    }

    // end the transaction (thereby freeing the lock on the table)
    dbCmd = "END";
    dbt->ExecuteSQL(dbCmd, resultSetPtr);
    PQclear(resultSetPtr);

    resultStatus.code = COPY_SUCCEEDED;
    resultStatus.message = "Success";
    return resultStatus;
}

novadaq::Status NamedConfigUtils::
copySubsystemConfig(const IDSpec& sourceId, bool makeParentChildLink,
                    const SimpleParameterSet& valuesToOverride,
                    std::string baselineTimestamp,
                    bool overrideIntermediateUpdate,
                    int64_t& resultId,
                    const std::string& subsystemNamesTableFile)
{
    novadaq::Status resultStatus;
    PGresult* resultSetPtr;
    boost::shared_ptr<NOVADB::Table> dbt;
    if (sourceId.isSlowControlsConfiguration() &&
        subsystemNamesTableFile == DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::
                  Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(subsystemNamesTableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    else {
        dbUser = novadaq::EnvVarCache::getInstance().getEnvVar("USER");
    }
    dbt->GetConnection();

    // look up the parameters associated with the source configuration
    NamedConfigSelectionCriteria selections;
    selections[ID_COLUMN_NAME] =
        boost::lexical_cast<std::string>(sourceId.getConfigId());
    NamedConfigDetailsPtr sourceCfgDetails = _getDetails(dbt, selections);
    if (sourceCfgDetails.get() == 0) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "The lookup of the input subsystem ";
        resultStatus.message.append("configuration ID failed.");
        return resultStatus;
    }
    std::string sourceConfigName;
    if (! getFieldValue(sourceCfgDetails, NAME_COLUMN_NAME,
                        sourceConfigName)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the configuration name ";
        resultStatus.message.append("for the source configuration.");
        return resultStatus;
    }
    std::string sourceDcmOpMode;
    if (! getFieldValue(sourceCfgDetails, DCMOPMODE_COLUMN_NAME,
                        sourceDcmOpMode)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the DCM operating mode ";
        resultStatus.message.append("for the source configuration.");
        return resultStatus;
    }
    std::string sourceDetectorName;
    if (! getFieldValue(sourceCfgDetails, DETECTORNAME_COLUMN_NAME,
                        sourceDetectorName)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the detector name ";
        resultStatus.message.append("for the source configuration.");
        return resultStatus;
    }
    std::string sourceSubsystemName;
    if (! getFieldValue(sourceCfgDetails, SUBSYSTEMNAME_COLUMN_NAME,
                        sourceSubsystemName)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the subsystem name ";
        resultStatus.message.append("for the source configuration.");
        return resultStatus;
    }
    std::string sourceConfigType;
    if (! getFieldValue(sourceCfgDetails, CONFIGTYPE_COLUMN_NAME,
                        sourceConfigType)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the configuration ";
        resultStatus.message.append("type for the source configuration.");
        return resultStatus;
    }
    std::string sourceTimestamp;
    if (! getFieldValue(sourceCfgDetails, INSERTTIME_COLUMN_NAME,
                        sourceTimestamp)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to determine the timestamp for the ";
        resultStatus.message.append("source configuration.");
        return resultStatus;
    }

    // determine the values of the parameters that we'll use to identify
    // the target configuration
    bool targetAndSourceHaveSameKey = true;
    std::string targetConfigName;
    if (valuesToOverride.get(NAME_COLUMN_NAME, targetConfigName)) {
        if (targetConfigName != sourceConfigName) {
            targetAndSourceHaveSameKey = false;
        }
    }
    else {
        targetConfigName = sourceConfigName;
    }
    std::string targetDcmOpMode;
    if (valuesToOverride.get(DCMOPMODE_COLUMN_NAME, targetDcmOpMode)) {
        if (targetDcmOpMode != sourceDcmOpMode) {
            targetAndSourceHaveSameKey = false;
        }
    }
    else {
        targetDcmOpMode = sourceDcmOpMode;
    }
    std::string targetDetectorName;
    if (valuesToOverride.get(DETECTORNAME_COLUMN_NAME, targetDetectorName)) {
        if (targetDetectorName != sourceDetectorName) {
            targetAndSourceHaveSameKey = false;
        }
    }
    else {
        targetDetectorName = sourceDetectorName;
    }
    std::string targetSubsystemName;
    if (valuesToOverride.get(SUBSYSTEMNAME_COLUMN_NAME, targetSubsystemName)) {
        if (targetSubsystemName != sourceSubsystemName) {
            targetAndSourceHaveSameKey = false;
        }
    }
    else {
        targetSubsystemName = sourceSubsystemName;
    }
    std::string targetConfigType;
    if (valuesToOverride.get(CONFIGTYPE_COLUMN_NAME, targetConfigType)) {
        if (targetConfigType != sourceConfigType) {
            targetAndSourceHaveSameKey = false;
        }
    }
    else {
        targetConfigType = sourceConfigType;
    }

    // lock the table
    std::string dbCmd = "BEGIN";
    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(dbCmd, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_COMMAND_OK) {
        PQclear(resultSetPtr);

        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to begin the database transaction.";
        return resultStatus;
    }
    PQclear(resultSetPtr);
    dbCmd = "LOCK TABLE ";
    dbCmd.append(dbt->Schema() + "." + dbt->Name());
    dbCmd.append(" IN ACCESS EXCLUSIVE MODE");
    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(dbCmd, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_COMMAND_OK) {
        PQclear(resultSetPtr);

        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to lock DB table ";
        resultStatus.message.append(dbt->Schema() + "." + dbt->Name());
        resultStatus.message.append(".");
        return resultStatus;
    }
    PQclear(resultSetPtr);

    // check if the latest instance of the target configuration is changing
    std::string queryString = "SELECT ";
    queryString.append(ISCHANGING_COLUMN_NAME);
    queryString.append(", ");
    queryString.append(ID_COLUMN_NAME);
    queryString.append(" FROM ");
    queryString.append(dbt->Schema() + "." + dbt->Name());
    queryString.append(" WHERE ");
    queryString.append(NAME_COLUMN_NAME);
    queryString.append("='");
    queryString.append(targetConfigName);
    queryString.append("' AND ");
    queryString.append(DCMOPMODE_COLUMN_NAME);
    queryString.append("='");
    queryString.append(targetDcmOpMode);
    queryString.append("' AND ");
    queryString.append(DETECTORNAME_COLUMN_NAME);
    queryString.append("='");
    queryString.append(targetDetectorName);
    queryString.append("' AND ");
    queryString.append(SUBSYSTEMNAME_COLUMN_NAME);
    queryString.append("='");
    queryString.append(targetSubsystemName);
    queryString.append("' AND ");
    queryString.append(CONFIGTYPE_COLUMN_NAME);
    queryString.append("='");
    queryString.append(targetConfigType);
    queryString.append("' ORDER BY ");
    queryString.append(ID_COLUMN_NAME);
    queryString.append(" DESC LIMIT 1");

    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(queryString, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_TUPLES_OK) {
        PQclear(resultSetPtr);

        resultStatus.code = COPY_FAILED;
        resultStatus.message = "Unable to check if the latest instance ";
        resultStatus.message.append("of the target configuration ");
        resultStatus.message.append("is changing.");
        return resultStatus;
    }
    if (PQntuples(resultSetPtr) > 0) {
        std::string resultString = PQgetvalue(resultSetPtr, 0, 0);
        boost::to_lower(resultString);
        if (resultString == "t") {
            resultStatus.code = COPY_FOUND_UPDATE_IN_PROGRESS;
            resultStatus.message = "Concurrent changes are in progress ";
            resultStatus.message.append("(subsystemConfigId = ");
            resultStatus.message.append(PQgetvalue(resultSetPtr, 0, 1));
            resultStatus.message.append(").");

            PQclear(resultSetPtr);

            dbCmd = "END";
            dbt->ExecuteSQL(dbCmd, resultSetPtr);

            return resultStatus;
        }
    }
    PQclear(resultSetPtr);

    // check if there are intermediate changes to the target configuration
    if (! overrideIntermediateUpdate) {
        boost::trim(baselineTimestamp);
        if (baselineTimestamp.length() == 0) {
            if (targetAndSourceHaveSameKey) {
                baselineTimestamp = sourceTimestamp;
            }
            else {
                baselineTimestamp = NOVADB::Util::GetFarPastTimeAsString();
            }
        }

        queryString = "SELECT ";
        queryString.append(ID_COLUMN_NAME);
        queryString.append(" FROM ");
        queryString.append(dbt->Schema() + "." + dbt->Name());
        queryString.append(" WHERE ");
        queryString.append(NAME_COLUMN_NAME);
        queryString.append("='");
        queryString.append(targetConfigName);
        queryString.append("' AND ");
        queryString.append(DCMOPMODE_COLUMN_NAME);
        queryString.append("='");
        queryString.append(targetDcmOpMode);
        queryString.append("' AND ");
        queryString.append(DETECTORNAME_COLUMN_NAME);
        queryString.append("='");
        queryString.append(targetDetectorName);
        queryString.append("' AND ");
        queryString.append(SUBSYSTEMNAME_COLUMN_NAME);
        queryString.append("='");
        queryString.append(targetSubsystemName);
        queryString.append("' AND ");
        queryString.append(CONFIGTYPE_COLUMN_NAME);
        queryString.append("='");
        queryString.append(targetConfigType);
        queryString.append("' AND ");
        queryString.append(INSERTTIME_COLUMN_NAME);
        queryString.append(">'");
        queryString.append(baselineTimestamp);
        queryString.append("' ORDER BY ");
        queryString.append(ID_COLUMN_NAME);
        queryString.append(" DESC");

        resultSetPtr = 0;
        if (! dbt->ExecuteSQL(queryString, resultSetPtr) ||
            PQresultStatus(resultSetPtr) != PGRES_TUPLES_OK) {
            PQclear(resultSetPtr);

            resultStatus.code = COPY_FAILED;
            resultStatus.message = "Unable to check if there are ";
            resultStatus.message.append("intermediate changes in the ");
            resultStatus.message.append("target configuration.");
            return resultStatus;
        }
        if (PQntuples(resultSetPtr) > 0) {
            resultStatus.code = COPY_FOUND_INTERMEDIATE_UPDATE;
            resultStatus.message = "Found intermediate changes in the ";
            resultStatus.message.append("target configuration ");
            resultStatus.message.append("(e.g. subsystemConfigId = ");
            resultStatus.message.append(PQgetvalue(resultSetPtr, 0, 0));
            resultStatus.message.append(").");

            PQclear(resultSetPtr);

            dbCmd = "END";
            dbt->ExecuteSQL(dbCmd, resultSetPtr);

            return resultStatus;
        }
        PQclear(resultSetPtr);
    }

    // add the new target row
    const NOVADB::Column* colPtr;
    std::string columnNamesClause;
    std::string valuesClause;
    std::vector<std::string> columnNames = dbt->GetColNames();
    for (uint32_t idx = 0; idx < columnNames.size(); ++idx) {

        std::string columnName = boost::to_lower_copy(columnNames[idx]);
        if (columnName == NamedConfigUtils::ID_COLUMN_NAME) {continue;}
        if (columnName == NamedConfigUtils::USAGECOUNT_COLUMN_NAME) {continue;}
        if (columnName == NamedConfigUtils::UPDATETIME_COLUMN_NAME) {continue;}
        if (columnName == NamedConfigUtils::UPDATEUSER_COLUMN_NAME) {continue;}

        if (columnName == NamedConfigUtils::ISCHANGING_COLUMN_NAME) {
            if (columnNamesClause.length() > 0) {
                columnNamesClause.append(",");
                valuesClause.append(",");
            }
            columnNamesClause.append(columnName);
            valuesClause.append("true");
            continue;
        }

        if (columnName == NamedConfigUtils::ISCOMPLETE_COLUMN_NAME) {
            if (columnNamesClause.length() > 0) {
                columnNamesClause.append(",");
                valuesClause.append(",");
            }
            columnNamesClause.append(columnName);
            valuesClause.append("false");
            continue;
        }

        if (columnName == NamedConfigUtils::PARENTID_COLUMN_NAME) {
            if (makeParentChildLink) {
                if (columnNamesClause.length() > 0) {
                    columnNamesClause.append(",");
                    valuesClause.append(",");
                }
                columnNamesClause.append(columnName);
                valuesClause.append(boost::lexical_cast<std::string>
                                    (sourceId.getConfigId()));
            }
            continue;
        }

        if (columnName == NamedConfigUtils::INSERTTIME_COLUMN_NAME) {
            if (columnNamesClause.length() > 0) {
                columnNamesClause.append(",");
                valuesClause.append(",");
            }
            columnNamesClause.append(columnName);
            valuesClause.append("'");
            valuesClause.append(NOVADB::Util::GetCurrentTimeAsString());
            valuesClause.append("'");
            continue;
        }

        if (columnName == NamedConfigUtils::INSERTUSER_COLUMN_NAME) {
            if (columnNamesClause.length() > 0) {
                columnNamesClause.append(",");
                valuesClause.append(",");
            }
            columnNamesClause.append(columnName);
            valuesClause.append("'");
            valuesClause.append(dbUser);
            valuesClause.append("'");
            continue;
        }

        std::string value;
        if (! valuesToOverride.get(columnName, value) && 
            ! getFieldValue(sourceCfgDetails, columnName, value)) {
            resultStatus.code = COPY_FAILED;
            resultStatus.message = "The lookup of the value for the ";
            resultStatus.message.append(columnName);
            resultStatus.message.append(" column failed.");
            return resultStatus;
        }

        colPtr = dbt->GetCol(columnName);
        std::string columnType = "string";
        if (colPtr != 0) {
            columnType = colPtr->Type();
        }

        if (columnNamesClause.length() > 0) {
            columnNamesClause.append(",");
            valuesClause.append(",");
        }
        columnNamesClause.append(columnName);
        if (columnType == "string" || columnType == "timestamp") {
            valuesClause.append("'");
            valuesClause.append(value);
            valuesClause.append("'");
        }
        else if (columnType == "bool") {
            try {
                int intValue = boost::lexical_cast<int>(value);
                if (intValue) {
                    valuesClause.append("true");
                }
                else {
                    valuesClause.append("false");
                }
            }
            catch (...) {
                valuesClause.append(value);
            }
        }
        else {
            valuesClause.append(value);
        }
    }

    std::string insertString = "INSERT INTO ";
    insertString.append(dbt->Schema() + "." + dbt->Name());
    insertString.append(" (");
    insertString.append(columnNamesClause);
    insertString.append(") VALUES (");
    insertString.append(valuesClause);
    insertString.append(")");

    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(insertString, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_COMMAND_OK) {
        PQclear(resultSetPtr);

        resultStatus.code = COPY_FAILED;
        resultStatus.message = "The writing of the new subsystem ";
        resultStatus.message.append("configuration into the database failed.");
        return resultStatus;
    }
    PQclear(resultSetPtr);
    if (! dbt->GetCurrSeqVal(NamedConfigUtils::ID_COLUMN_NAME, resultId)) {
        resultStatus.code = COPY_FAILED;
        resultStatus.message = "The lookup of the new subsystem ";
        resultStatus.message.append("configuration ID failed.");
        return resultStatus;
    }

    // end the transaction (thereby freeing the lock on the table)
    dbCmd = "END";
    dbt->ExecuteSQL(dbCmd, resultSetPtr);
    PQclear(resultSetPtr);

    resultStatus.code = COPY_SUCCEEDED;
    resultStatus.message = "Success";
    return resultStatus;
}

bool NamedConfigUtils::
markConfigDone(const IDSpec& idSpec, bool alsoComplete,
               const std::string& tableFile)
{
    PGresult* resultSetPtr;
    const NOVADB::Column* colPtr;
    boost::shared_ptr<NOVADB::Table> dbt;
    if (tableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        if (idSpec.isGlobalConfiguration()) {
            if (idSpec.isDAQConfiguration()) {
                dbt.reset(new NOVADB::
                          Table(DEFAULT_GLOBAL_NAMES_TABLE_FILE.c_str()));
            }
            else {  // slow controls
                dbt.reset(new NOVADB::
                          Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
            }
        }
        else {  // subsystem
            if (idSpec.isDAQConfiguration()) {
                dbt.reset(new NOVADB::
                          Table(DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
            }
            else {  // slow controls
                dbt.reset(new NOVADB::
                          Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
            }
        }
    }
    else {
        dbt.reset(new NOVADB::Table(tableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    else {
        dbUser = novadaq::EnvVarCache::getInstance().getEnvVar("USER");
    }
    dbt->GetConnection();

    // lock the table
    std::string dbCmd = "BEGIN";
    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(dbCmd, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_COMMAND_OK) {
        PQclear(resultSetPtr);
        return false;
    }
    PQclear(resultSetPtr);
    dbCmd = "LOCK TABLE ";
    dbCmd.append(dbt->Schema() + "." + dbt->Name());
    dbCmd.append(" IN ACCESS EXCLUSIVE MODE");
    resultSetPtr = 0;
    if (! dbt->ExecuteSQL(dbCmd, resultSetPtr) ||
        PQresultStatus(resultSetPtr) != PGRES_COMMAND_OK) {
        PQclear(resultSetPtr);
        return false;
    }
    PQclear(resultSetPtr);

    // update the flag(s)
    std::string updateString = "UPDATE ";
    updateString.append(dbt->Schema() + "." + dbt->Name());
    updateString.append(" SET ");
    updateString.append(ISCHANGING_COLUMN_NAME);
    updateString.append("=false");
    if (alsoComplete) {
        updateString.append(", ");
        updateString.append(ISCOMPLETE_COLUMN_NAME);
        updateString.append("=true");
    }

    colPtr = dbt->GetCol(UPDATETIME_COLUMN_NAME.c_str());
    if (colPtr != 0) {
        updateString.append(", ");
        updateString.append(UPDATETIME_COLUMN_NAME);
        updateString.append("='");
        updateString.append(NOVADB::Util::GetCurrentTimeAsString());
        updateString.append("'");
    }

    colPtr = dbt->GetCol(UPDATEUSER_COLUMN_NAME.c_str());
    if (colPtr != 0) {
        updateString.append(", ");
        updateString.append(UPDATEUSER_COLUMN_NAME);
        updateString.append("='");
        updateString.append(dbUser);
        updateString.append("'");
    }

    updateString.append(" WHERE ");
    updateString.append(ID_COLUMN_NAME);
    updateString.append("=");
    updateString.append(boost::lexical_cast<std::string>
                        (idSpec.getConfigId()));

    resultSetPtr = 0;
    bool status = dbt->ExecuteSQL(updateString, resultSetPtr);
    PQclear(resultSetPtr);

    // end the transaction (thereby freeing the lock on the table)
    dbCmd = "END";
    dbt->ExecuteSQL(dbCmd, resultSetPtr);
    PQclear(resultSetPtr);

    return status;
}

bool NamedConfigUtils::
incrementUsageCount(const IDSpec& idSpec, const std::string& tableFile)
{
    boost::shared_ptr<NOVADB::Table> dbt;
    if (tableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        if (idSpec.isGlobalConfiguration()) {
            if (idSpec.isDAQConfiguration()) {
                dbt.reset(new NOVADB::
                          Table(DEFAULT_GLOBAL_NAMES_TABLE_FILE.c_str()));
            }
            else {  // slow controls
                dbt.reset(new NOVADB::
                          Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
            }
        }
        else {  // subsystem
            if (idSpec.isDAQConfiguration()) {
                dbt.reset(new NOVADB::
                          Table(DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
            }
            else {  // slow controls
                dbt.reset(new NOVADB::
                          Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
            }
        }
    }
    else {
        dbt.reset(new NOVADB::Table(tableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    dbt->Clear();
    dbt->SetValidityRange(ID_COLUMN_NAME, idSpec.getConfigId());
    if (! dbt->LoadFromDB()) {return false;}
    if (dbt->NRow() != 1) {return false;}

    int usageCount = 0;
    NOVADB::Row* rowPtr = dbt->GetRow(0);
    NOVADB::Column* colPtr = rowPtr->Col(USAGECOUNT_COLUMN_NAME);
    if (! colPtr->IsNull()) {
        if (! colPtr->Get(usageCount)) {return false;}
    }
    ++usageCount;

    rowPtr->Update(USAGECOUNT_COLUMN_NAME, usageCount);
    if (! dbt->WriteToDB()) {return false;}

    return true;
}

/**
 * Fetches the full set of details for the specified named
 * configuration *and* all of its ancestors.
 *
 * The details of the requested configuration are first in the returned
 * list, followed by its parent then grandparent, etc.  If the
 * requested configuration has no ancestors, then only the details of
 * the requested configuration will be returned in the list.
 *
 * NOTE that this method assumes certain column names for the database
 * table that contains the configuration names.  If any of these
 * assumptions are not valid for the given globalNamesTableFile,
 * then this method will always return an empty list.
 *
 * @param globalNameSpec The specification of the named global
 *        configuration of interest.
 * @param globalNamesTableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 *
 * @returns the list of configuration details.
 */
std::vector<NamedConfigUtils::NamedConfigDetailsPtr> NamedConfigUtils::
getGlobalConfigDetailAncestry(const GlobalNameSpec& globalNameSpec,
                              const std::string& globalNamesTableFile)
{
    std::vector<NamedConfigDetailsPtr> emptyList;
    int64_t globalConfigId;

    boost::shared_ptr<NOVADB::Table> dbt;
    if (globalNameSpec.isSlowControlsConfiguration() &&
        globalNamesTableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        dbt.reset(new NOVADB::Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
    }
    else {
        dbt.reset(new NOVADB::Table(globalNamesTableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    NamedConfigSelectionCriteria selections;
    selections[NAME_COLUMN_NAME] = globalNameSpec.getConfigName();
    selections[DCMOPMODE_COLUMN_NAME] = globalNameSpec.getDCMOpMode();
    selections[DETECTORID_COLUMN_NAME] = globalNameSpec.getDetectorId();
    if (globalNameSpec.hasTimestamp()) {
        selections[INSERTTIME_COLUMN_NAME] = globalNameSpec.getTimestamp();
    }
    if (! globalNameSpec.skipInvalidStateTest()) {
        selections[ISOBSOLETE_COLUMN_NAME] = "false";
        selections[ISCHANGING_COLUMN_NAME] = "false";
        selections[ISCOMPLETE_COLUMN_NAME] = "true";
    }

    if (! _getId(dbt, selections, globalConfigId)) {
        return emptyList;
    }

    IDSpec globalIdSpec(globalConfigId, IDSpec::GLOBAL,
                        (globalNameSpec.isSlowControlsConfiguration() ?
                         IDSpec::SLOWCONTROLS : IDSpec::DAQ));
    return getConfigDetailAncestry(globalIdSpec, globalNamesTableFile);
}

/**
 * Fetches the full set of details for the specified named
 * configuration *and* all of its ancestors.
 *
 * The details of the requested configuration are first in the returned
 * list, followed by its parent then grandparent, etc.  If the
 * requested configuration has no ancestors, then only the details of
 * the requested configuration will be returned in the list.
 *
 * NOTE that this method assumes certain column names for the database
 * table that contains the configuration names.  If any of these
 * assumptions are not valid for the tableFile that is given (or 
 * (inferred), then this method will always return an empty list.
 *
 * @param idSpec The ID-based specification of the named configuration
 *        of interest.
 * @param tableFile The name of the XML file (with any
 *        needed path information) that contains the definition of the
 *        database table that contains the named configurations.
 *
 * @returns the list of configuration details.
 */
std::vector<NamedConfigUtils::NamedConfigDetailsPtr> NamedConfigUtils::
getConfigDetailAncestry(const IDSpec& idSpec,
                        const std::string& tableFile)
{
    // I need to revisit this method to make it more efficient - KAB

    std::vector<NamedConfigDetailsPtr> emptyList;
    std::vector<NamedConfigDetailsPtr> detailsList;
    std::vector<int64_t> idList;

    boost::shared_ptr<NOVADB::Table> dbt;
    if (tableFile == DEFAULT_GLOBAL_NAMES_TABLE_FILE) {
        if (idSpec.isGlobalConfiguration()) {
            if (idSpec.isDAQConfiguration()) {
                dbt.reset(new NOVADB::
                          Table(DEFAULT_GLOBAL_NAMES_TABLE_FILE.c_str()));
            }
            else {  // slow controls
                dbt.reset(new NOVADB::
                          Table(DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE.c_str()));
            }
        }
        else {  // subsystem
            if (idSpec.isDAQConfiguration()) {
                dbt.reset(new NOVADB::
                          Table(DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
            }
            else {  // slow controls
                dbt.reset(new NOVADB::
                          Table(DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE.c_str()));
            }
        }
    }
    else {
        dbt.reset(new NOVADB::Table(tableFile.c_str()));
    }
    std::string daqEnv = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(daqEnv);
    dbt->SetTableType(NOVADB::kDAQTable);
    std::string dbUser =
        novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
        dbt->SetUser(dbUser);
    }
    dbt->GetConnection();

    idList.push_back(idSpec.getConfigId());
    uint32_t previousSize = idList.size() - 1;

    while (idList.size() > previousSize) {
        previousSize = idList.size();
        if (! _findNextAncestor(dbt, idList)) {
            return emptyList;
        }
    }

    for (uint32_t cfgIdx = 0; cfgIdx < idList.size(); ++cfgIdx) {
        dbt->Clear();
        dbt->SetValidityRange(ID_COLUMN_NAME, idList[cfgIdx]);
        if (! dbt->LoadFromDB() || dbt->NRow() != 1) {
            return emptyList;
        }
        NamedConfigDetailsPtr detailsPtr(new NamedConfigDetails());
        std::string columnValue;
        NOVADB::Row* rowPtr = dbt->GetRow(0);
        for (int idx = 0; idx < dbt->NCol(); ++idx) {
            NOVADB::Column& colRef = rowPtr->Col(idx);
            if (colRef.IsNull()) {
                (*detailsPtr)[colRef.Name()] = "";
            }
            else if (colRef.Get(columnValue)) {
                (*detailsPtr)[colRef.Name()] = columnValue;
            }
            else {
                return emptyList;
            }
        }
        detailsList.push_back(detailsPtr);
    }
    return detailsList;
}

/**
 * Returns the specified field from the details collection.  If the
 * requested field is not found, this method returns false.
 */
bool NamedConfigUtils::
getFieldValue(const NamedConfigDetailsPtr& details,
              const std::string& fieldName,
              std::string& fieldValue)
{
    if (details.get() == 0) {return false;}
    NamedConfigDetails::const_iterator iter = details->find(fieldName);
    if (iter != details->end()) {
        fieldValue = iter->second;
        return true;
    }
    return false;
}

/**
 * Fetches the column values for the specified selection criteria.
 *
 * @param dbt A shared pointer to the database table that contains
 *        the named configurations.
 * @param selections The selection criteria.
 *
 * @returns an empty pointer if something failed, otherwise valid details
 */
NamedConfigUtils::NamedConfigDetailsPtr NamedConfigUtils::
_getDetails(const boost::shared_ptr<NOVADB::Table>& dbt,
            const NamedConfigSelectionCriteria& selections)
{
    NamedConfigDetailsPtr detailsPtr;
    dbt->Clear();

    NamedConfigSelectionCriteria::const_iterator iter;
    NamedConfigSelectionCriteria::const_iterator iterEnd =
        selections.end();
    for (iter = selections.begin(); iter != iterEnd; ++iter) {
        std::string columnName = iter->first;
        std::string lowBound = iter->second;
        std::string highBound = iter->second;
        if (columnName == INSERTTIME_COLUMN_NAME) {
            lowBound = NOVADB::Util::GetFarPastTimeAsString();
        }
        if (! dbt->SetValidityRange(columnName, lowBound, highBound)) {
            return detailsPtr;
        }
    }

    // explicitly request that later entries be returned first
    if (! dbt->AddOrderColumn(ID_COLUMN_NAME)) {return detailsPtr;}
    dbt->SetOrderDesc();
    //dbt->SetVerbosity(1);
    if (dbt->LoadFromDB() && dbt->NRow() > 0) {
        NamedConfigDetailsPtr tmpDetailsPtr(new NamedConfigDetails());
        std::string columnValue;
        NOVADB::Row* rowPtr = dbt->GetRow(0);
        for (int idx = 0; idx < dbt->NCol(); ++idx) {
            NOVADB::Column& colRef = rowPtr->Col(idx);
            if (colRef.IsNull()) {
                (*tmpDetailsPtr)[colRef.Name()] = "";
            }
            else if (colRef.Get(columnValue)) {
                (*tmpDetailsPtr)[colRef.Name()] = columnValue;
            }
            else {
                return detailsPtr;
            }
        }
        detailsPtr = tmpDetailsPtr;
    }

    return detailsPtr;
}

/**
 * Fetches the database ID for the specified selection criteria.
 *
 * @param dbt A shared pointer to the database table that contains
 *        the named configurations.
 * @param selections The selection criteria.
 * @param id The database ID of the named configuration [output parameter].
 *        (If the status returned by this method is false, then the value
 *        returned in this parameter is undefined.)
 *
 * @returns true if the lookup of the ID of the specified named
 *        configuration worked, false if not.
 */
bool NamedConfigUtils::
_getId(const boost::shared_ptr<NOVADB::Table>& dbt,
       const NamedConfigSelectionCriteria& selections,
       int64_t& id)
{
    NamedConfigDetailsPtr detailsPtr = _getDetails(dbt, selections);

    std::string tmpString;
    if (! getFieldValue(detailsPtr, ID_COLUMN_NAME, tmpString)) {return false;}

    try {id = boost::lexical_cast<int64_t>(tmpString);}
    catch (...) {return false;}

    return true;
}

/**
 * Finds the next ancestor in the specified list of named configuration
 * IDs and adds it to the list.
 *
 * If there is no ancestor for the latest ID in the list, this method
 * returns true, but the idList is unchanged.
 *
 * @param dbt A shared pointer to the database table that contains
 *        the named configurations.
 * @param idList The list of database IDs for the set of
 *        named configurations [input/output parameter].
 *        (If the status returned by this method is false, then the
 *        values returned in this parameter are undefined.)
 *
 * @returns true if the lookup of the next ancestor ID worked, false if not.
 */
bool NamedConfigUtils::
_findNextAncestor(const boost::shared_ptr<NOVADB::Table>& dbt,
                  std::vector<int64_t>& idList)
{
    if (idList.size() < 1) {return false;}

    dbt->Clear();
    int64_t oldestId = idList[idList.size()-1];
    dbt->SetValidityRange(ID_COLUMN_NAME, oldestId);
    if (dbt->LoadFromDB()) {
        if (dbt->NRow() == 1) {
            NOVADB::Row* rowPtr = dbt->GetRow(0);
            NOVADB::Column* colPtr = rowPtr->Col(PARENTID_COLUMN_NAME);
            if (colPtr->IsNull()) {
                return true;
            }
            int64_t parentId;
            if (colPtr->Get(parentId)) {
                idList.push_back(parentId);
                return true;
            }
        }
    }
    return false;
}

/**
 * Finds the next child in the specified list of named configuration
 * IDs and adds it to the list.
 *
 * If there is no child for the latest ID in the list, this method
 * returns true, but the idList is unchanged.
 *
 * @param dbt A shared pointer to the database table that contains
 *        the named configurations.
 * @param idList The list of database IDs for the set of
 *        named configurations [input/output parameter].
 *        (If the status returned by this method is false, then the
 *        values returned in this parameter are undefined.)
 *
 * @returns true if the lookup of the next child ID worked, false if not.
 */
bool NamedConfigUtils::
_findNextChild(const boost::shared_ptr<NOVADB::Table>& dbt,
               std::vector<int64_t>& idList)
{
    if (idList.size() < 1) {return false;}

    dbt->Clear();
    int64_t newestId = idList[idList.size()-1];
    dbt->SetValidityRange(PARENTID_COLUMN_NAME, newestId);
    dbt->SetValidityRange(ISOBSOLETE_COLUMN_NAME, false);
    dbt->SetValidityRange(ISCHANGING_COLUMN_NAME, false);
    dbt->SetValidityRange(ISCOMPLETE_COLUMN_NAME, true);
    if (! dbt->LoadFromDB()) {return false;}

    if (dbt->NRow() == 0) {return true;}

    NOVADB::Row* rowPtr = dbt->GetRow(0);
    NOVADB::Column* colPtr = rowPtr->Col(ID_COLUMN_NAME);
    int64_t thisId;
    if (! colPtr->Get(thisId)) {return false;}

    idList.push_back(thisId);
    return true;
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
