#include <DatabaseUtils/DAQConfig/SubsystemNameSpec.h>
#include <boost/lexical_cast.hpp>

namespace dbutils {
namespace daqconfig {

const std::string SubsystemNameSpec::DAQMONITOR_SUBSYSTEM_NAME("DaqMonitor");
const std::string SubsystemNameSpec::DCM_SUBSYSTEM_NAME("DCMApplication");
const std::string SubsystemNameSpec::DL_SUBSYSTEM_NAME("DataLogger");
const std::string SubsystemNameSpec::GT_SUBSYSTEM_NAME("GlobalTrigger");
const std::string SubsystemNameSpec::SM_SUBSYSTEM_NAME("SimulationManager");
const std::string SubsystemNameSpec::DDT_SUBSYSTEM_NAME("DDTManager");
const std::string SubsystemNameSpec::TDUMANAGER_SUBSYSTEM_NAME("TDUManager");
const std::string SubsystemNameSpec::BNEVB_SUBSYSTEM_NAME("BufferNodeEVB");
const std::string SubsystemNameSpec::CONNECT_CONFIG_TYPE("connections");
const std::string SubsystemNameSpec::HARDWARE_CONFIG_TYPE("hardware");
const std::string SubsystemNameSpec::RUN_CONFIG_TYPE("run");

/**
 * Creates a subsystem name specification object from the specified
 * configuration name, DCM operating mode, detector ID, subsystem name,
 * and configuration type.
 *
 * @param configName The named subsystem configuration of interest.
 * @param dcmOperatingMode The operating mode of interest (e.g. DCS or DSO).
 * @param detectorId The detector of interest (e.g. NDOS or FarDet).
 * @param subsystemName The name of the subsystem of interest (e.g.
 *        DCMApplication or GlobalTrigger).
 * @param configType The type of the configuration data (e.g. run
 *        or hardware).
 */
SubsystemNameSpec::
SubsystemNameSpec(const std::string& configName,
                  const std::string& dcmOperatingMode,
                  const std::string& detectorId,
                  const std::string& subsystemName,
                  const std::string& configType) :
    _configName(configName), _dcmOperatingMode(dcmOperatingMode),
    _detectorId(detectorId), _subsystemName(subsystemName),
    _configType(configType), _hasTimestamp(false),
    _isSlowControlsConfig(false), _usageCount(0), _sortIndex(-1)
{
}

/**
 * Creates a subsystem name specification object from the specified
 * configuration name, DCM operating mode, detector ID, subsystem name,
 * configuration type, and timestamp.
 *
 * @param configName The named subsystem configuration of interest.
 * @param dcmOperatingMode The operating mode of interest (e.g. DCS or DSO).
 * @param detectorId The detector of interest (e.g. NDOS or FarDet).
 * @param subsystemName The name of the subsystem of interest (e.g.
 *        DCMApplication or GlobalTrigger).
 * @param configType The type of the configuration data (e.g. run
 *        or hardware).
 * @param timestamp The date/time string that should be used to
 *        determine the time-based validity context for the named
 *        configuration.  Supported formats include "yyyy-mm-dd HH:MM:SS".
 */
SubsystemNameSpec::
SubsystemNameSpec(const std::string& configName,
                  const std::string& dcmOperatingMode,
                  const std::string& detectorId,
                  const std::string& subsystemName,
                  const std::string& configType,
                  const std::string& timestamp) :
    _configName(configName), _dcmOperatingMode(dcmOperatingMode),
    _detectorId(detectorId), _subsystemName(subsystemName),
    _configType(configType), _hasTimestamp(true), _timestamp(timestamp),
    _isSlowControlsConfig(false), _usageCount(0), _sortIndex(-1)
{
}

/**
 * Creates a string representation of the name specification.
 */
std::string SubsystemNameSpec::toString(int overallIndent) const
{
    std::string workString;
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("Named Subsystem Configuration:\n");
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  configuration name: " + _configName + "\n");
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  DCM operating mode: " + _dcmOperatingMode + "\n");
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  detector name: " + _detectorId + "\n");
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  subsystem name: " + _subsystemName + "\n");
    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  configuration type: " + _configType + "\n");

    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    if (_hasTimestamp) {
        workString.append("  timestamp: " + _timestamp);
    }
    else {
        workString.append("  no timestamp");
    }
    workString.append("\n");

    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  usage count: " +
                      boost::lexical_cast<std::string>(_usageCount) + "\n");

    if (overallIndent > 0) {_addSpaces(workString, overallIndent);}
    workString.append("  DAQ or Slow Controls: ");
    workString.append((_isSlowControlsConfig ? "Slow Controls" : "DAQ"));

    return workString;
}

/**
 * Tests if this instance is less than the specified one.
 */
bool SubsystemNameSpec::
operator<(const SubsystemNameSpec& other) const
{
    if (this->_sortIndex > -1 && other._sortIndex > -1) {
        return (this->_sortIndex < other._sortIndex);
    }

    if (this->_usageCount > other._usageCount) {return false;}
    if (this->_usageCount < other._usageCount) {return true;}
    // else _usageCount is equal

    if (this->_detectorId > other._detectorId) {return false;}
    if (this->_detectorId < other._detectorId) {return true;}
    // else _detectorId is equal

    if (this->_subsystemName > other._subsystemName) {return false;}
    if (this->_subsystemName < other._subsystemName) {return true;}
    // else _subsystemName is equal

    if (this->_configType > other._configType) {return false;}
    if (this->_configType < other._configType) {return true;}
    // else _configType is equal

    if (this->_configName > other._configName) {return false;}
    if (this->_configName < other._configName) {return true;}
    // else _configName is equal

    if (this->_dcmOperatingMode > other._dcmOperatingMode) {return false;}
    if (this->_dcmOperatingMode < other._dcmOperatingMode) {return true;}
    // else _dcmOperatingMode is equal

    if (this->_timestamp > other._timestamp) {return false;}
    if (this->_timestamp < other._timestamp) {return true;}
    // else _timestamp is equal

    return false;
}

/**
 * Adds a specified number of spaces to a string.
 */
void SubsystemNameSpec::_addSpaces(std::string& workString, int count) const
{
    for (int idx = 0; idx < count; ++idx) {
        workString.append(" ");
    }
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
