/**
 * NdmcClient implementation file
 * $Id: NdmcClient.cpp,v 1.21 2013/06/13 15:27:04 kasahara Exp $
 */
#include <typeinfo>
#include <iostream>
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "NovaDAQMonitorClient/NdmcClient.h"
#include "NovaDAQMonitorClient/NdmcMetric.h"

using namespace NDMC_NAMESPACE;

/**
 * Constructor
 */
NdmcClient::NdmcClient(int32_t partition, std::string appName, std::string groupName, 
                       uint32_t sendTimeInterval) :
                _partition(partition), _applicationName(appName), _groupName(""),
                _userGroupName(groupName), _sendTimeInterval(sendTimeInterval), _lastSendTime(0),
                _clientThread(NULL) {

  _updateGroupName();

  LOG_DEBUG("NdmcClient") << "NdmcClient::ctor @" << this << " w/ partition " << getPartition()
                          << ", app name " << getApplicationName() << ", group name " 
                          << getGroupName() << ", send time interval " 
                          << getSendTimeInterval() <<  "(s)." << std::endl;

  _clientThread = new boost::thread(boost::bind(&novadaq::Runnable::run,this));
 
}

/**
 * Destructor
 */
NdmcClient::~NdmcClient() {

  LOG_DEBUG("NdmcClient") << "NdmcClient::dtor @" << this << "." << std::endl;

  stop();

  if ( _clientThread != NULL ) {
    _clientThread -> join();
    delete _clientThread;
  }

  // All adopted metrics are owned
  std::map<int32_t,NdmcMetricABC*>::iterator itr;
  for ( itr = _metricMap.begin(); itr != _metricMap.end(); itr++ ) {
    if ( itr -> second != NULL ) delete (itr -> second);
  }
  _metricMap.clear();

}


/**
 * Set partition
 */
bool NdmcClient::setPartition(int32_t partition) {

  boost::mutex::scoped_lock lock(_mutex);

  _partition = partition;
  _updateGroupName();
  return true;

}

/**
 * Set application name
 */
void NdmcClient::setApplicationName(std::string appName) {

  boost::mutex::scoped_lock lock(_mutex);

  _applicationName = appName;
  _updateGroupName();

}

/**
 * Set group name
 */
void NdmcClient::setGroupName(std::string groupName) {

  boost::mutex::scoped_lock lock(_mutex);

  _userGroupName = groupName;
  _updateGroupName();

}

/**
 * Update group name to <appname>+P<partition> if groupname hasn't been set,
 * or <groupname>+P<partition> if it has been set.
 * Update managed metrics with current partition, appname, and groupname.
 */
void NdmcClient::_updateGroupName() {

  std::ostringstream oss;
  if ( !_userGroupName.empty() ) oss << _userGroupName;
  else if ( !getApplicationName().empty() ) oss << getApplicationName();
  if ( getPartition() >= 0 ) {
    if ( oss.str().empty() ) oss << "P" << getPartition();
    else oss << " P" << getPartition();
  }
  _groupName = oss.str();

  if ( _metricMap.size() == 0 ) return;

  std::ostringstream oss2;
  std::map<int32_t, NdmcMetricABC*>::iterator itr;
  int im = 0;
  for ( itr = _metricMap.begin(); itr != _metricMap.end(); itr++, im++ ) {
    (itr -> second) -> setPartition(getPartition());
    (itr -> second) -> setApplicationName(getApplicationName());
    (itr -> second) -> setGroupName(getGroupName());
    if ( im > 0 ) oss2 << ", ";
    if ( im % 5 == 0 ) oss2 << "\n";
    oss2 << (itr -> second) -> getName();
  }

  ::mf::LogInfo("NdmcClient") << "NdmcClient::_updataGroupName - Partition " << getPartition() 
			      << ", AppName \"" << getApplicationName() << "\", GroupName \""
                              << getGroupName() << "\" applied to metric(s): " << oss2.str() 
                              << "." << std::endl;

}

/**
 * Main process loop. run() will run continuously until _stopRequested is set with a stop request. 
 */
void NdmcClient::run() {

  LOG_DEBUG("NdmcClient") << "NdmcClient::run." << std::endl;
  
  _running = true; // base class Runnable data member.  Allows isRunning() to be used.
  while ( !_stopRequested ) { // stop requested by invoking stop method
    sleep(1);
    if ( _timeToSend() ) {
      send(); // send the metrics
    }
  }

  _running = false;
 
  LOG_DEBUG("NdmcClient") << "NdmcClient::run - done." << std::endl;

}

/**
 * Set metric send time interval.
 */
void NdmcClient::setSendTimeInterval(uint32_t sendTimeInterval) {

  boost::mutex::scoped_lock lock(_mutex);
  _sendTimeInterval = sendTimeInterval;
  
}

/**
 * Adopt a metric to own
 * @param metric - pointer to user created NdmcMetric
 * @return -1 if already metric of this name, else metricId indexing metric in list
 */
int32_t NdmcClient::_adoptMetric(NdmcMetricABC* metric) {

  int32_t metricID = -1;
  std::string metricName = metric -> getName();
  std::map<int32_t, NdmcMetricABC*>::const_iterator citr;
  for ( citr = _metricMap.begin(); citr != _metricMap.end(); citr++ ) {
    if ( metric -> getName() == (citr -> second) -> getName() ) {
      ::mf::LogWarning("NdmcClient") << "NdmcClient::_adoptMetric - invoked for metric w/name " 
             << metricName << " failed. Duplicate name with existing entry." << std::endl;
      return -1;
    }
    if ( citr -> first > metricID ) metricID = citr -> first;
  }

  metricID += 1;
  LOG_DEBUG("NdmcClient") << "NdmcClient::_adoptMetric - adopting metric with name " 
                          << metricName << ", assigned metric ID " << metricID << "." 
                          << std::endl;
  _metricMap[metricID] = metric;
  return metricID;

}

/**
 * Destroy metric by ID.
 */
void NdmcClient::destroyMetric(int32_t metricID) {
 
  LOG_DEBUG("NdmcClient") << "NdmcClient::destroyMetric - Metric ID " << metricID << "." 
                          << std::endl;

  boost::mutex::scoped_lock lock(_mutex);
   
  std::map<int32_t, NdmcMetricABC*>::iterator itr = _metricMap.find(metricID);
  if ( itr != _metricMap.end() ) {
    if ( itr->second != NULL ) delete (itr->second);
    _metricMap.erase(itr);
  }
 
}

/**
 * Destroy all managed metrics and remove from map
 */
void NdmcClient::destroyMetrics() {

  LOG_DEBUG("NdmcClient") << "NdmcClient::destroyMetrics." << std::endl;

  boost::mutex::scoped_lock lock(_mutex);

  // All adopted metrics are owned
  std::map<int32_t,NdmcMetricABC*>::iterator itr;
  for ( itr = _metricMap.begin(); itr != _metricMap.end(); itr++ ) {
    if ( itr -> second != NULL ) delete (itr -> second);
  }
  _metricMap.clear();

}

/**
 * Get a metric by metric ID
 *
 * @param metricID indexing metric of interest
 * @return pointer to metric or NULL if doesn't exist 
 */
NdmcMetricABC* NdmcClient::_getMetric(int32_t metricID) const {
  
  std::map<int32_t, NdmcMetricABC*>::const_iterator citr = _metricMap.find(metricID);
  if ( citr == _metricMap.end() ) return NULL;
  return citr -> second;

}

/**
 * Return number of metrics
 *
 * @return number of metrics
 */
uint32_t NdmcClient::getNumMetric() {

  boost::mutex::scoped_lock lock(_mutex);
  return _metricMap.size();

}

/**
 * Send all metrics.  Return number of metrics successfully sent.  Compare
 * to NdmcClient::getNumMetric().
 * @return number of metrics successfully sent
 */
int32_t NdmcClient::send() {

  boost::mutex::scoped_lock lock(_mutex);

  int32_t nokay = 0;

  std::map<int32_t, NdmcMetricABC*>::iterator itr;
  for ( itr = _metricMap.begin(); itr != _metricMap.end(); itr++ ) {
    if ( (itr->second) -> send() ) nokay++;
  }
  return nokay;

}

/**
 * Private method to determine if its time to send the next batch of metrics
 */
bool NdmcClient::_timeToSend() {

  struct tms ctime;
  clock_t wall = times(&ctime);
  if ( _lastSendTime == 0 ) {
    _lastSendTime = wall;
    return false;
  }
  double deltaw = ((double)(wall-_lastSendTime))*10000./CLOCKS_PER_SEC;

  if ( deltaw >= _sendTimeInterval ) {
    _lastSendTime = wall;
    return true;
  }

  return false;  

}
