#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsRuntimeException.h>
#include <unistd.h>
#include <boost/lexical_cast.hpp>

namespace gov {

namespace fnal {

namespace cd {

namespace rms {

namespace provider {

int DDSConnection::_instanceCount = 0;
boost::mutex DDSConnection::_staticMutex;

/**
 * Creates a new DDSConnection given the application name and partition.
 *
 * @param applicationName The name of the application generating/receiving
 * messages.
 * @param applicationPartition The partition of the application that is
 * generating/receiving messages.  
 */
DDSConnection::DDSConnection(const std::string applicationName,
                             const int partitionNumber) {
    _applicationName = applicationName;
    _partitionNumber = partitionNumber;

    char hostname[80];
    memset(hostname, 0, 80);
    if (gethostname(hostname, 79) == 0) {
        _uuidGenerator = new util::UUIDGenerator(hostname);
        std::string tmpString(hostname);
        _myHostName = tmpString;
    }
    else {
        _uuidGenerator = new util::UUIDGenerator("C++DDSConnection");
        _myHostName = "Unknown";
    }

    _debugFlag = 0;
    char *debugEnvVar = getenv("RMS_DEBUG");
    if (debugEnvVar != NULL) {
        std::string debugString(debugEnvVar);
        try {
            _debugFlag = boost::lexical_cast<int>(debugString);
        }
        catch (boost::bad_lexical_cast &excpt) {
            std::cerr << "*** ERROR: Invalid RMS_DEBUG value: \""
                      << debugString << "\"." << std::endl;
        }
    }

    boost::mutex::scoped_lock sl(_staticMutex);
    ++_instanceCount;
    if (_instanceCount == 1) {
        try {
           DPSingleton::Instance().connect_participant();
        }
        catch (...) {
            std::cerr << "Unable to connect DomainParticipant in "
                      << "DDSConnection constructor." << std::endl;
        }
    }
}

/**
 * Destructor.
 */
DDSConnection::~DDSConnection() {
    close();

    boost::mutex::scoped_lock sl(_staticMutex);
    --_instanceCount;
    if (_instanceCount == 0) {
        DPSingleton::Instance().disconnect_participant();
    }
}

/**
 * Determine if the given destination can be reached with the DDS provider.
 *
 * @param dest The destination to check to see if messages can be sent to it.
 * @return True if the destination can be reached, false otherwise.
 */
bool DDSConnection::supportsDestination(const base::RmsDestination& dest) {
    return true;
}

/**
 * Closes the connection.  After calling this method, the
 * connection should no longer be used since it will no
 * longer be connected to the underlying provider system.
 */
void DDSConnection::close() {
    if (_uuidGenerator) {
        delete _uuidGenerator;
        _uuidGenerator = NULL;
    }
}

};  // end of namespace provider

};  // end of namespace rms

};  // end of namespace cd

};  // end of namespace fnal

};  // end of namespace gov
